// ############################################################################################
//  VegetationFractals.cc
//  Implementation of various CIndefiniteAreaFractal derivative classes
//  specialized to imitate plants and leaves
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the algorithms used in this file were written based on the book :
//    Fractal Vision : Put Fractals to Work for You
//    Dick Oliver
//    Sams Publishing
//    1992
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "VegetationFractals.h"


namespace vge {

//
// Bamboo 1
//
FractalBamboo::FractalBamboo()
{
  if(SEED_REPLICATOR_FRACTAL_ALLOC(1, 4, 3)) {
    SEED_POINT(0, 0, -10,   0);
    SEED_POINT(0, 1,  -5, 150);
    SEED_POINT(0, 2,   5, 150);
    SEED_POINT(0, 3,  10,   0);
    SEED_REPLICA(0, 50, 50,  700,  700, -5, 20); // Left
    SEED_REPLICA(1, 50, 50, -700, -700,  5, 20); // Right
    SEED_REPLICA(2, 80, 80,    0,    0,  0, 60); // Top
  }
}

bool FractalBamboo::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  if(un32CurLevel <= 1)
    SetColor(28, 24, 0);
  else {
    uint32_t un32C = uint32_t((31.0f / un32MaxLevel) * un32CurLevel);
    un32C = (un32C > 31) ? 31 : un32C;
    SetColor(31 - un32C, un32C, 0);
  }
  return(true);
}

bool FractalBamboo::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32MaxLevel, un32CurLevel, un32TotRelPart, un32CurRelPart);
  SetColor(0, uint32_t(_rnFloat.GeneratePositive(21) + 10), 0);
  return(true);
}


//
// Fern 1
//
FractalFern::FractalFern()
{
  if(SEED_REPLICATOR_FRACTAL_ALLOC(1, 2, 3)) {
    SEED_POINT(0, 0, 0,   0);
    SEED_POINT(0, 1, 0, 150);
    SEED_REPLICA(0, 40, 40,  1100,  1100, 1,  50); // Left
    SEED_REPLICA(1, 40, 40, -1100, -1100, 1,  50); // Right
    SEED_REPLICA(2, 70, 70,   -30,   -30, 1, 100); // Top
  }
}

bool FractalFern::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32TotPart, un32CurPart, un32TotRelPart, un32CurRelPart);
  uint32_t un32C = uint32_t((21.0f / un32MaxLevel) * un32CurLevel) + 10;
  un32C = (un32C > 31) ? 31 : un32C;
  SetColor(0, un32C, 0);
  return(true);
}

bool FractalFern::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32MaxLevel, un32CurLevel, un32TotRelPart, un32CurRelPart);
  SetColor(0, uint32_t(_rnFloat.GeneratePositive(21) + 10), 0);
  return(true);
}


//
// Maple Leaf 1
//
FractalMapleLeaf::FractalMapleLeaf()
{
  if(SEED_REPLICATOR_FRACTAL_ALLOC(1, 8, 4)) {
    SEED_POINT(0, 0,   0,  60);
    SEED_POINT(0, 1,  15,  15);
    SEED_POINT(0, 2,  40,  25);
    SEED_POINT(0, 3,  35,   0);
    SEED_POINT(0, 4,   0, -15);
    SEED_POINT(0, 5, -35,   0);
    SEED_POINT(0, 6, -40,  25);
    SEED_POINT(0, 7, -15,  15);
    SEED_REPLICA(0, 65, 65,  600,  600, -9,  1); // Left
    SEED_REPLICA(1, 65, 65, -600, -600,  9,  1); // Right
    SEED_REPLICA(2, 52, 52,    0,    0,  0, 30); // Top
    SEED_REPLICA(3, 15, 50,    0,    0,  0, -8); // Bottom
  }
}

bool FractalMapleLeaf::CosmosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotPart, uint32_t un32CurPart, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_3(un32TotPart, un32CurPart, un32TotRelPart);
  uint32_t un32C = uint32_t((31.0f / un32MaxLevel) * un32CurLevel);
  if(un32CurRelPart == 3)
    SetColor(0, un32C >> 2, 0);
  else if(un32CurRelPart == 2)
    SetColor(0, un32C >> 1, 0);
  else
    SetColor(0, un32C, 0);
  return(true);
}

bool FractalMapleLeaf::ChaosUpdateColor(uint32_t un32MaxLevel, uint32_t un32CurLevel, uint32_t un32TotRelPart, uint32_t un32CurRelPart)
{
  UNUSED_PARAMS_4(un32MaxLevel, un32CurLevel, un32TotRelPart, un32CurRelPart);
  SetColor(0, uint32_t(_rnFloat.GeneratePositive(18) + 13), 0);
  return(true);
}

} // namespace vge
