/**
 * \file FoggedCeilingLineSegment.h
 * <hr>
 * \internal
 * \todo While implementing the SIMD optimization, try to remove the __dummy*
 *       variables as many as possible.
 * <hr>
 */
// ############################################################################################
//  FoggedCeilingLineSegment.h
//  Definition and impelementation of fogged ceiling-line-segment classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _3D_ENVIRONMENT___FOGGED_CEILING_LINE_SEGMENT_H_
#define _3D_ENVIRONMENT___FOGGED_CEILING_LINE_SEGMENT_H_

// VortexGE headers
#include "3DEnvironment.h"
#include "Miscellaneous3DStructs.h"


namespace vge {

  //!
  //! \internal
  //! \brief The fogged-untextured-flat-shaded ceiling-line-segment class
  //!
  //! \api pretty stable
  class _CFogUFCLineSeg : VCLASS_UNCOPYABLE
  {
    public:
      _CFogUFCLineSeg();
      // Ceil line functions
      void Init(const _TPoint2D &tP1, const _TPoint2D &tP2);
      void ClipTop(const int32_t n32Top);
      inline void Step();
      inline int32_t GetX()      const { return(_n32X >> _XSTEP_PREC_); }
      inline int32_t GetY()      const { return(_n32Y); }
      inline int32_t GetZ()      const { return(_n32Z); }
      inline int32_t GetF()      const { return(_n32F); }
      inline int32_t GetHeight() const { return(_n32EdgeHeight); }

    protected:
      int32_t _n32X1, _n32Y1, _n32X2, _n32Y2;
#ifdef __USE_AVEC__
      // NOTE: Do not change the order !!!
      int32_t _sse _n32X;
      int32_t      _n32Y, _n32Z, _n32F;
      int32_t      _n32XStep, _n32YStep, _n32ZStep, _n32FStep;
#else
      int32_t _n32X, _n32Y, _n32Z, _n32F;
      int32_t _n32XStep, _n32ZStep, _n32FStep;
#endif
      int32_t _n32EdgeHeight;
  };

  void _CFogUFCLineSeg::Step()
  {
#ifdef __USE_AVEC__
          v4si &c_xyzf = I32_TO_V4SI(_n32X);
    const v4si &s_xyzf = I32_TO_V4SI(_n32XStep);
    c_xyzf += s_xyzf;
#else
    _n32X += _n32XStep;
    _n32Y++;
    _n32Z += _n32ZStep;
    _n32F += _n32FStep;
#endif
    _n32EdgeHeight--;
  }


  //!
  //! \internal
  //! \brief The fogged-untextured-gouraud-shaded ceiling-line-segment class
  //!
  //! \api pretty stable
  class _CFogUGCLineSeg : VCLASS_UNCOPYABLE
  {
    public:
      _CFogUGCLineSeg();
      // Ceil line functions
      void Init(const _TPoint2D &tP1, const _TPoint2D &tP2);
      void ClipTop(const int32_t n32Top);
      inline void Step();
      inline int32_t GetX()      const { return(_n32X >> _XSTEP_PREC_); }
      inline int32_t GetY()      const { return(_n32Y); }
      inline int32_t GetZ()      const { return(_n32Z); }
      inline int32_t GetF()      const { return(_n32F); }
      inline int32_t GetR()      const { return(_n32R); }
      inline int32_t GetG()      const { return(_n32G); }
      inline int32_t GetB()      const { return(_n32B); }
      inline int32_t GetHeight() const { return(_n32EdgeHeight); }

    protected:
      int32_t _n32X1, _n32Y1, _n32X2, _n32Y2;
#ifdef __USE_AVEC__
      // NOTE: Do not change the order !!!
      int32_t _sse _n32X;
      int32_t      _n32Y, _n32Z, _n32F;
      int32_t      _n32R, _n32G, _n32B, ___dummy1;
      int32_t      _n32XStep, _n32YStep, _n32ZStep, _n32FStep;
      int32_t      _n32RStep, _n32GStep, _n32BStep, ___dummy2;
#else
      int32_t _n32X, _n32Y, _n32Z, _n32F;
      int32_t _n32R, _n32G, _n32B;
      int32_t _n32XStep, _n32ZStep, _n32FStep;
      int32_t _n32RStep, _n32GStep, _n32BStep;
#endif
      int32_t _n32EdgeHeight;
  };

  void _CFogUGCLineSeg::Step()
  {
#ifdef __USE_AVEC__
          v4si &c_xyzf = I32_TO_V4SI(_n32X);
          v4si &c_rgb_ = I32_TO_V4SI(_n32R);
    const v4si &s_xyzf = I32_TO_V4SI(_n32XStep);
    const v4si &s_rgb_ = I32_TO_V4SI(_n32RStep);
    c_xyzf += s_xyzf;
    c_rgb_ += s_rgb_;
#else
    _n32X += _n32XStep;
    _n32Y++;
    _n32Z += _n32ZStep;
    _n32F += _n32FStep;
    _n32R += _n32RStep;
    _n32G += _n32GStep;
    _n32B += _n32BStep;
#endif
    _n32EdgeHeight--;
  }


  //!
  //! \internal
  //! \brief The fogged-textured-flat-shaded ceiling-line-segment class
  //!
  //! \api pretty stable
  class _CFogTFCLineSeg : VCLASS_UNCOPYABLE
  {
    public:
      _CFogTFCLineSeg();
      // Ceil line functions
      void Init(const _TPoint2D &tP1, const _TPoint2D &tP2);
      void ClipTop(const int32_t n32Top);
      inline void Step();
      inline int32_t GetX()      const { return(_n32X >> _XSTEP_PREC_); }
      inline int32_t GetY()      const { return(_n32Y); }
      inline int32_t GetZ()      const { return(_n32Z); }
      inline int32_t GetF()      const { return(_n32F); }
      inline int32_t GetU()      const { return(_n32U); }
      inline int32_t GetV()      const { return(_n32V); }
      inline int32_t GetHeight() const { return(_n32EdgeHeight); }

    protected:
      int32_t _n32X1, _n32Y1, _n32X2, _n32Y2;
#ifdef __USE_AVEC__
      // NOTE: Do not change the order !!!
      int32_t _sse _n32X;
      int32_t      _n32Y, _n32Z, _n32F;
      int32_t      _n32U, _n32V, ___dummy1, ___dummy2;
      int32_t      _n32XStep, _n32YStep, _n32ZStep, _n32FStep;
      int32_t      _n32UStep, _n32VStep, ___dummy3, ___dummy4;
#else
      int32_t _n32X, _n32Y, _n32Z, _n32F;
      int32_t _n32U, _n32V;
      int32_t _n32XStep, _n32ZStep, _n32FStep;
      int32_t _n32UStep, _n32VStep;
#endif
      int32_t _n32EdgeHeight;
  };

  void _CFogTFCLineSeg::Step()
  {
#ifdef __USE_AVEC__
          v4si &c_xyzf = I32_TO_V4SI(_n32X);
          v4si &c_uv__ = I32_TO_V4SI(_n32U);
    const v4si &s_xyzf = I32_TO_V4SI(_n32XStep);
    const v4si &s_uv__ = I32_TO_V4SI(_n32UStep);
    c_xyzf += s_xyzf;
    c_uv__ += s_uv__;
#else
    _n32X += _n32XStep;
    _n32Y++;
    _n32Z += _n32ZStep;
    _n32F += _n32FStep;
    _n32U += _n32UStep;
    _n32V += _n32VStep;
#endif
    _n32EdgeHeight--;
  }


  //!
  //! \internal
  //! \brief The fogged-textured-gouraud-shaded ceiling-line-segment class
  //!
  //! \api pretty stable
  class _CFogTGCLineSeg : VCLASS_UNCOPYABLE
  {
    public:
      _CFogTGCLineSeg();
      // Ceil line functions
      void Init(const _TPoint2D &tP1, const _TPoint2D &tP2);
      void ClipTop(const int32_t n32Top);
      inline void Step();
      inline int32_t GetX()      const { return(_n32X >> _XSTEP_PREC_); }
      inline int32_t GetY()      const { return(_n32Y); }
      inline int32_t GetZ()      const { return(_n32Z); }
      inline int32_t GetF()      const { return(_n32F); }
      inline int32_t GetU()      const { return(_n32U); }
      inline int32_t GetV()      const { return(_n32V); }
      inline int32_t GetR()      const { return(_n32R); }
      inline int32_t GetG()      const { return(_n32G); }
      inline int32_t GetB()      const { return(_n32B); }
      inline int32_t GetHeight() const { return(_n32EdgeHeight); }

    protected:
      int32_t _n32X1, _n32Y1, _n32X2, _n32Y2;
#ifdef __USE_AVEC__
      // NOTE: Do not change the order !!!
      int32_t _sse _n32X;
      int32_t      _n32Y, _n32Z, _n32F;
      int32_t      _n32U, _n32V, ___dummy1, ___dummy2;
      int32_t      _n32R, _n32G, _n32B, ___dummy3;
      int32_t      _n32XStep, _n32YStep, _n32ZStep, _n32FStep;
      int32_t      _n32UStep, _n32VStep, ___dummy4, ___dummy5;
      int32_t      _n32RStep, _n32GStep, _n32BStep, ___dummy6;
#else
      int32_t _n32X, _n32Y, _n32Z, _n32F;
      int32_t _n32U, _n32V;
      int32_t _n32R, _n32G, _n32B;
      int32_t _n32XStep, _n32ZStep, _n32FStep;
      int32_t _n32UStep, _n32VStep;
      int32_t _n32RStep, _n32GStep, _n32BStep;
#endif
      int32_t _n32EdgeHeight;

  };

  void _CFogTGCLineSeg::Step()
  {
#ifdef __USE_AVEC__
          v4si &c_xyzf = I32_TO_V4SI(_n32X);
          v4si &c_uv__ = I32_TO_V4SI(_n32U);
          v4si &c_rgb_ = I32_TO_V4SI(_n32R);
    const v4si &s_xyzf = I32_TO_V4SI(_n32XStep);
    const v4si &s_uv__ = I32_TO_V4SI(_n32UStep);
    const v4si &s_rgb_ = I32_TO_V4SI(_n32RStep);
    c_xyzf += s_xyzf;
    c_uv__ += s_uv__;
    c_rgb_ += s_rgb_;
#else
    _n32X += _n32XStep;
    _n32Y++;
    _n32Z += _n32ZStep;
    _n32F += _n32FStep;
    _n32U += _n32UStep;
    _n32V += _n32VStep;
    _n32R += _n32RStep;
    _n32G += _n32GStep;
    _n32B += _n32BStep;
#endif
    _n32EdgeHeight--;
  }

} // namespace vge
#endif
