/** \file Exception.h */
// ############################################################################################
//  Exception.h
//  Definition & implementation of VortexGE exception classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _BASE___EXCEPTION_H_
#define _BASE___EXCEPTION_H_

// VortexGE headers
#include "VBasics.h"


namespace vge {

  //!
  //! \brief Error codes
  //!
  //! Note on usage:
  //! - Invoking VError() using one of these error codes, will throw the
  //!   appropriate exception class.
  //! - Invoking VFatalError() with any of these codes will always throw
  //!   CFatalError exception class, regardless to which class the error
  //!   code is actually belongs to.
  //! - Some error codes defined here are not used yet (they may become
  //!   used in the future). Those error codes are :
  //!     - VERR_None
  //!     - VERR_Unknown
  //!     - VERR_AlreadyInitialized
  //!     - VERR_SeekStream
  //!     - VERR_WriteDevice
  //!     - VERR_PackData
  //!     - VERR_CompressData
  //!     - VERR_Busy
  //!     - VERR_Loaded
  //!     - VERR_NotLoaded
  //! - Some error codes are no longer used by the library and have been
  //!   deprecated. They will be removed soon. Those error codes are :
  //!     - VERR_FATAL_InvalidUsage
  //!     - VERR_MultipleInstance
  //!     - VERR_XGetWindowProperty
  //!     - VERR_MsgRcv
  //!     - VERR_InvalidEnvVar
  //!
  //! \api pretty stable
  enum EERRCode {

    /*** Unknown Errors ************************************/
    VERR_None, VERR_Unknown,
    /************************************ Unknown Errors ***/

    /*** Fatal Errors **************************************/ _S_FATAL_ERRORS_,
    VERR_FATAL_GotFatalSignal, VERR_FATAL_IOError,
    VERR_FATAL_InvalidUsage,
    /************************************** Fatal Errors ***/ _E_FATAL_ERRORS_,

    /*** Logic Errors **************************************/ _S_LOGIC_ERRORS_,
    VERR_NotDefined,       VERR_NotImplemented,
    VERR_InvalidCall,      VERR_InvalidParameter,
    VERR_NotInitialized,   VERR_AlreadyInitialized,
    VERR_MultipleInstance,
    /************************************** Logic Errors ***/ _E_LOGIC_ERRORS_,

    /*** Runtime Errors ************************************/ _S_RTIME_ERRORS_,
    //
    /* XWindow Errors ------------------------------------ */ _S_XWIND_ERRORS_,
    VERR_XOpenDisplay,       VERR_XVideoMode,
    VERR_XGetWindowProperty, VERR_XCreateGC,
    VERR_XOpenOM,            VERR_XOpenIM,
    VERR_XOpenIC,
    VERR_XCreateImage,       VERR_XShmAttach,
    VERR_XLoadQueryFont,     VERR_XCreateFontSet,
    VERR_XCreateFontCursor,  VERR_XGrabPointer,
    /* --------------------------------------------------- */ _E_XWIND_ERRORS_,
    //
    /* System Errors ------------------------------------- */ _S_SYSTS_ERRORS_,
    VERR_SYS_mkdir,    VERR_SYS_rmdir,
    VERR_SYS_chdir,    VERR_SYS_getcwd,
    VERR_SYS_link,     VERR_SYS_unlink,
    VERR_SYS_stat,
    VERR_MsgGet,       VERR_MsgCtl,
    VERR_MsgSnd,       VERR_MsgRcv,
    VERR_ShmGet,       VERR_ShmAt,
    VERR_CreateThread, VERR_LockMutex,
    VERR_UnlockMutex,  VERR_MutexNotOwned,
    /* --------------------------------------------------- */ _E_SYSTS_ERRORS_,
    //
    /* IO Errors ----------------------------------------- */ _S_INOUT_ERRORS_,
    VERR_OpenDirectory,   VERR_OpenFile,
    VERR_OpenFromStream,  VERR_SeekStream,
    VERR_ReadStream,      VERR_WriteStream,
    VERR_OpenDevice,      VERR_DeviceNotOpenend,
    VERR_GetDeviceParam,  VERR_SetDeviceParam,
    VERR_ReadDevice,      VERR_WriteDevice,
    /* --------------------------------------------------- */ _E_INOUT_ERRORS_,
    //
    /* Scripting (VortexLua) ----------------------------- */ _S_VXLUA_ERRORS_,
    VERR_CompileScript, VERR_ExecuteScript,
    /* --------------------------------------------------- */ _E_VXLUA_ERRORS_,
    //
    /* Other Errors -------------------------------------- */
    //
    // Validity
    VERR_InvalidEnvVar,     VERR_InvalidStream,
    VERR_InvalidHeaderData, VERR_IncorrectVersion,
    //
    // Data Packing & Compression
    VERR_PackData,     VERR_UnpackData,
    VERR_CompressData, VERR_UncompressData,
    //
    // Miscellaneous State
    VERR_Busy,     VERR_InUse,        VERR_NotResponding,
    VERR_NotFound, VERR_NotSupported, VERR_NotEnoughMemory,
    VERR_Loaded,   VERR_NotLoaded,    VERR_TooMany,
    VERR_Exist,    VERR_NotExist,     VERR_Timeout,
    /* --------------------------------------------------- */
    //
    /************************************ Runtime Errors ***/ _E_RTIME_ERRORS_

  };


  //!
  //! \brief Source information class
  //!
  //! \api pretty stable
  class VGE_EXPORT CSourceInfo {
    public:
      //!
      //! Construct a source information object based on the given parameters
      CSourceInfo(const char *pcFunction, const char *pcFile, int32_t n32Line);
      //!
      //! Copy ctor.
      CSourceInfo(const CSourceInfo &csSInfo);
      //!
      //! Standard dtor.
      virtual ~CSourceInfo() throw();

      // Returns some descriptions of the source location
      //!
      //! Check if the CSourceInfo instance contains usable/valid informations.
      virtual bool Valid() const { return(_n32Line >= 0); }
      //!
      //! Returns the function name.
      virtual const char *Function() const { return(_csFunction.c_str()); }
      //!
      //! Returns the file name.
      virtual const char *File() const { return(_csFile.c_str()); }
      //!
      //! Returns the line number
      virtual int32_t Line() const { return(_n32Line); }

      // Just for convenicence
      virtual CSourceInfo &operator=(const CSourceInfo &csSInfo);

    protected:
      string  _csFunction;
      string  _csFile;
      int32_t _n32Line;
  };


  //!
  //! \brief Generic exception class
  //!
  //! \api pretty stable
  class VGE_EXPORT CException : public exception {
    public:
      // Construct an exception object based on the given parameters
      CException(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CException(const CException &csException);
      //!
      //! Standard dtor.
      virtual ~CException() throw();

      // Returns some descriptions of the exception
      //!
      //! Returns the error description.
      virtual const char *What() const throw() { return(_csWhat.c_str()); }
      //!
      //! Returns the error location.
      virtual const CSourceInfo &Location() const throw() { return(_csLocation); }
      //!
      //! Returns the error code.
      virtual EERRCode Code() const throw() { return(_eCode); }

      // Just for convenicence
      virtual CException &operator=(const CException &csException);

    protected:
      // From std::exception
      virtual const char *what() const throw() { return(What()); }

      // Data
      string      _csWhat;
      CSourceInfo _csLocation;
      EERRCode    _eCode;
  };


  //!
  //! \brief Generic exception class for fatal errors, will only be thrown
  //!        if invoking VFatalError()
  //!
  //! \api pretty stable
  class VGE_EXPORT CFatalError : public CException {
    public:
      //!
      //! Construct a CFatalError exception object based on the given parameters
      CFatalError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CFatalError(const CFatalError &csException);
      //!
      //! Standard dtor.
      virtual ~CFatalError() throw();
  };


  //!
  //! \brief Exception class for logic errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CLogicError : public CException {
    public:
      //!
      //! Construct a CLogicError exception object based on the given parameters
      CLogicError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CLogicError(const CLogicError &csException);
      //!
      //! Standard dtor.
      virtual ~CLogicError() throw();
  };


  //!
  //! \brief Exception class for runtime errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CRuntimeError : public CException {
    public:
      //!
      //! Construct a CRuntimeError exception object based on the given parameters
      CRuntimeError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CRuntimeError(const CRuntimeError &csException);
      //!
      //! Standard dtor.
      virtual ~CRuntimeError() throw();
  };


  //!
  //! \brief Exception class for XWindow errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CXWindowError : public CRuntimeError {
    public:
      //!
      //! Construct a CXWindowError exception object based on the given parameters
      CXWindowError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CXWindowError(const CXWindowError &csException);
      //!
      //! Standard dtor.
      virtual ~CXWindowError() throw();
  };



  //!
  //! \brief Exception class for system errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CSystemError : public CRuntimeError {
    public:
      //!
      //! Construct a CSystemError exception object based on the given parameters
      CSystemError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CSystemError(const CSystemError &csException);
      //!
      //! Standard dtor.
      virtual ~CSystemError() throw();
  };


  //!
  //! \brief Exception class for IO errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CIOError : public CRuntimeError {
    public:
      //!
      //! Construct a CIOError exception object based on the given parameters
      CIOError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CIOError(const CIOError &csException);
      //!
      //! Standard dtor.
      virtual ~CIOError() throw();
  };


  //!
  //! \brief Exception class for scripting (VortexLua) errors
  //!
  //! \api pretty stable
  class VGE_EXPORT CScriptingError : public CRuntimeError {
    public:
      //!
      //! Construct a CScriptingError exception object based on the given parameters
      CScriptingError(EERRCode eCode, const char *pcWhat, const CSourceInfo &csLocation);
      //!
      //! Copy ctor.
      CScriptingError(const CScriptingError &csException);
      //!
      //! Standard dtor.
      virtual ~CScriptingError() throw();
  };


  // For convenicence
  //!
  //! Throw the appropriate exception object based on the error code.
  void VGE_EXPORT VThrowExceptionBasedOnCode(EERRCode code, const char *msg, const CSourceInfo &loc) _attr_noreturn;

} // namespace vge
#endif
