// ############################################################################################
//  ApplicationObject.cc
//  Implementation of CBasicAppObjects class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "../AudioEnvironment/MID_Dummy/MDMYPlayer.h"
#include "../AudioEnvironment/MID_libTiMidity/MTIMPlayer.h"
#include "../AudioEnvironment/PCM_Dummy/ADMYSoundBuffer.h"
#include "../AudioEnvironment/PCM_OSS/AOSSSoundBuffer.h"
#include "../System/SystemUtilities.h"
#include "../VideoEnvironment/VideoEnvironment.h"
#include "ApplicationObject.h"


namespace vge {

//
// Messages to be displayed when failed opening audio device
//
static const char *failedOpeningAudioDeviceMessage(uint32_t textIndex)
{
  const char *str1 = "Couldn't open Audio Device";
  const char *str2 = "Couldn't open audio device \"%s\", it may be in use.\n\n"     \
                     "Terminate all other multimedia applications, or wait for a\n" \
                     "while until any sound server, such as \"artsd\" or \"esd\"\n" \
                     "suspends itself (or just terminate it).\n\n"                  \
                     "Click \"Yes\" to try again or \"No\" to disable audio ...\n";

  switch(textIndex) {
    case 0 : return(_M(str1));
    case 1 : return(_M(str2));
    default: return("");
  };
}


//
// A structure used for storing the global mouse settings
//
TMouseSettings TMouseSettings::GLOBAL_SET;


//
// A structure used for storing the global key settings
//
TKeySettings TKeySettings::GLOBAL_SET;


//
// The basic application objects class
//
char            CBasicAppObjects::_decSep = '.';
CMessageCatalog CBasicAppObjects::_msgCat;

CBasicAppObjects::CBasicAppObjects(const char *utf8WindowTitle, const char *pucWindowIconFileName,
                                   uint32_t    un32WindowWidth, uint32_t    un32WindowHeight,
                                   uint32_t    un32BufferWidth, uint32_t    un32BufferHeight,
                                   EAIM        eAudioInitMode,  E3DIM       e3DInitMode)
: _csAudioEnv    (CAudioEnv::AutoDetect()),
  _csMidiPlayer  (CMidiPlayer::AutoDetect()),
  _csVideoEnv    (NULL),
  _cs3DEnv       (NULL),
  _bAudioDisabled(false),
  _bMidiDisabled(false)
{
  _Initialize_(utf8WindowTitle, pucWindowIconFileName, un32WindowWidth, un32WindowHeight, un32BufferWidth, un32BufferHeight, eAudioInitMode, e3DInitMode);
}

CBasicAppObjects::CBasicAppObjects(ESDM        eDisplayMode,
                                   const char *utf8WindowTitle,
                                   const char *pucWindowIconFileName,
                                   EAIM        eAudioInitMode,
                                   E3DIM       e3DInitMode)
: _csAudioEnv    (CAudioEnv::AutoDetect()),
  _csMidiPlayer  (CMidiPlayer::AutoDetect()),
  _csVideoEnv    (NULL),
  _cs3DEnv       (NULL),
  _bAudioDisabled(false),
  _bMidiDisabled(false)
{
  uint32_t wW, wH, bW, bH;
  switch(eDisplayMode) {
    case ESDM_320X240_NORMAL : wW = 320; wH = 240; bW = 320; bH = 240; break;
    case ESDM_320X240_LOWRES : wW = 320; wH = 240; bW = 160; bH = 120; break;
    case ESDM_640X480_NORMAL : wW = 640; wH = 480; bW = 640; bH = 480; break;
    case ESDM_640X480_LOWRES : wW = 640; wH = 480; bW = 320; bH = 240; break;
    case ESDM_800X600_NORMAL : wW = 800; wH = 600; bW = 800; bH = 600; break;
    case ESDM_800X600_LOWRES : wW = 800; wH = 600; bW = 400; bH = 300; break;
    default                  : VError(VERR_InvalidParameter, _NFL_, "Invalid predefined display mode constant");
  };

  _Initialize_(utf8WindowTitle, pucWindowIconFileName, wW, wH, bW, bH, eAudioInitMode, e3DInitMode);
}

CBasicAppObjects::~CBasicAppObjects() throw()
{
  iVideoEnv().AutoRepeatOn();

  delete _cs3DEnv;
  delete _csVideoEnv;
}

const char *CBasicAppObjects::SystemLocale()
{
  static const char *s_LocVars[] = { "VGE_LOCALE",
                                     "LC_MESSAGES",
                                     "LC_ALL",
                                     "LANGUAGE",
                                     "LANG" };
  const char* plev;
  for(size_t i = 0; i < elemsize(s_LocVars); ++i) {
    if((plev = getenv(s_LocVars[i]))) return (plev);
  }
  return("C");
}

void CBasicAppObjects::InitializeI18N(const char *pathToMessageCatalog, const char *messagesLocale)
{
  // Initialize message catalog
  const char *env0   = getenv("VGE_MESSAGE_CATALOG_PATH");
  const char *envx   = env0 ? env0 : __LIB_HC_MCTLG_PATH__;
  const char *path = pathToMessageCatalog ? pathToMessageCatalog : envx;
  CBasicAppObjects::_msgCat.LoadCatalog(path, messagesLocale);

  // Get the character used for decimal separator
  char tmp[16];
  setlocale(LC_ALL, CBasicAppObjects::SystemLocale());
  snprintf(tmp, sizeof(tmp), "%3.1f", 1.0f);
  setlocale(LC_ALL, "en_US.UTF-8");
  CBasicAppObjects::_decSep = tmp[1];
}

void CBasicAppObjects::_Initialize_(const char *utf8WindowTitle, const char *pucWindowIconFileName,
                                    uint32_t    un32WindowWidth, uint32_t    un32WindowHeight,
                                    uint32_t    un32BufferWidth, uint32_t    un32BufferHeight,
                                    EAIM        eAudioInitMode,  E3DIM       e3DInitMode)
{
  // Instantiate the video environment
  _csVideoEnv = _nothrow_new CVideoEnv;
  if(!_csVideoEnv) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't instantiate the application's CVideoEnv instance");

  // Instantiate the 3D environment
  if(e3DInitMode == E3DIM_Disabled)
    VInfo(_NFL_, "3D renderer initialization is disabled by request");
  else {
    _cs3DEnv = _nothrow_new C3DEnv;
    if(!_cs3DEnv) {
      delete _csVideoEnv; _csVideoEnv = NULL;
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't instantiate the application's C3DEnv instance");
    }
  }
  if(e3DInitMode == E3DIM_Manual) VInfo(_NFL_, "3D renderer initialization is delayed by request (the user seems want to initialize it manually)");

  // Initialize the message catalog
  try {
    if(CBasicAppObjects::_msgCat.IsEmpty()) 
      CBasicAppObjects::InitializeI18N();
  }
  catch(...) {}

  // Make sure the paramaters are initialized with the correct values
  if(!un32WindowWidth)  un32WindowWidth  = 640;
  if(!un32WindowHeight) un32WindowHeight = 480;
  if(!utf8WindowTitle)  utf8WindowTitle  = _M("VortexGE Application Window");
  if(!un32BufferWidth)  un32BufferWidth  = un32WindowWidth;
  if(!un32BufferHeight) un32BufferHeight = un32WindowHeight;

  // Get the user preference on color palette
  bool uecm = false;
  char *env = getenv("VGE_VIDEO_CUSTOM_8BPP_PALETTE");
       if(VMatchEVar(env, "yes") == EMEV_Match) uecm = true;
  else if(VMatchEVar(env,  "no") == EMEV_Match) uecm = false;
  else if(env) VWarning(_NFL_, "Invalid value in environment variable \"VGE_VIDEO_CUSTOM_8BPP_PALETTE\"\nThe default system palette will be used");

  // Initialize video
  iVideoEnv().CreateMainWindow(un32WindowWidth, un32WindowHeight, utf8WindowTitle, pucWindowIconFileName, uecm);
  if(e3DInitMode == E3DIM_Normal) {
    iVideoEnv().CreateRenderImage(un32WindowWidth, un32WindowHeight);
    iVideoEnv().CreateRenderBuffer(un32BufferWidth, un32BufferHeight);
  }
  iVideoEnv().AutoRepeatOff();

  // Initialize 3D
  if(e3DInitMode == E3DIM_Normal) i3DEnv().Initialize(&iVideoEnv());

  // The user does not want audio
  if(eAudioInitMode == EAIM_Disabled) {
    _bAudioDisabled = true;
    VInfo(_NFL_, "Audio initialization is disabled by request");
  }
  // Initialize audio
  else {
    // Query for device name
    char acDevName[PATH_MAX];
    iAudioEnv().QueryDeviceName(acDevName, sizeof(acDevName));
    // Open the device
    for(;;) {
      // Try to open the device
      bool bOpened = false;
      try {
        iAudioEnv().OpenDevice(acDevName);
        bOpened = true;
      }
      catch(...) {
        // If the user specify that the audio must be enabled,
        // just rethrow the exception
        if(eAudioInitMode == EAIM_Enabled) throw;
      }
      // If opened, initialize and activate the device
      if(bOpened) {
        iAudioEnv().InitializeDevice();
        iAudioEnv().ActivateDevice();
        break;
      }
      // * If Failed, ask the user if he want to try again or disable the audio
      // * But, if "EAIM_NormalNoRetry" is specified then no message box will
      //   be displayed, just directly disable the audio.
      EMsgBoxRetVal eRet = EMBRV_Ignore;
      if(eAudioInitMode != EAIM_NormalNoRetry) {
        // Give the user a message box
        char acTmp[1024];
        snprintf(acTmp, sizeof(acTmp), failedOpeningAudioDeviceMessage(1), acDevName);
        eRet = iVideoEnv().MessageBox(failedOpeningAudioDeviceMessage(0), acTmp, EMBI_Warning, EMBB_YesNo, EMBS_Small);
      }
      // Check for the user choice
      if(eRet == EMBRV_No) {
        _bAudioDisabled = true;
        VWarning(_NFL_, "Audio is disabled because of initialization failure");
        break;
      }
    }
  }

  // Initialize midi (only if audio is not disabled)
  if(_bAudioDisabled)
    _bMidiDisabled = true;
  else {
    try {
      iMidiPlayer().InitializeDevice(&iAudioEnv());
    }
    catch(...) {
      if(eAudioInitMode == EAIM_Enabled) throw;
      _bMidiDisabled  = true;
      VWarning(_NFL_, "MIDI playback is disabled because of initialization failure");
    }
  }
}

} // namespace vge
