/** \file EventHandler.h */
// ############################################################################################
//  EventHandler.h
//  Definition & implementation of CEventHandler and CEventHandlerExt classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _HIGH_LEVEL_CLASS___EVENT_HANDLER_H_
#define _HIGH_LEVEL_CLASS___EVENT_HANDLER_H_

// VortexGE headers
#include "../VideoEnvironment/Clipboard.h"
#include "../VideoEnvironment/KeyButtonSymbol.h"
#include "FrameRateControl.h"


namespace vge {

  //!
  //! \brief The base of all event handler classess
  //!
  //! Shouldn't need to use this class directly. Use the derivative classes instead.
  //!
  //! \api beta
  class VGE_EXPORT CEventHandlerBase : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new CEventHandlerBase object using the given video environment.
      explicit CEventHandlerBase(CVideoEnv *pcsVidEnv);
      //!
      //! Standard dtor.
      virtual ~CEventHandlerBase() throw();

      // Hooking functions
      //!
      //! Hook this event handler class to the given video environment instance.
      inline void Hook(CVideoEnv *pcsVidEnv) { _pcsVidEnv = pcsVidEnv; };
      //!
      //! Returns the currently hooked video environment instance.
      inline CVideoEnv *HookedVideoEnv() const { return(_pcsVidEnv); };

      // Information functions
      //!
      //! Check if the application loop is ordered to exit or not.
      inline bool Exiting() const { return (_bDoExit); }

      // Overrideable loop functions
      //!
      //! This function must be overriden by user to implement the main application loop.
      virtual int32_t DoMainLoop() = 0;

      // Overrideable event handling functions
      //!
      //! This function must be overriden by user to implement the event pumping mechanism.
      virtual bool PumpEvents() = 0;
      //!
      //! This function must be overriden by user to implement the event handling mechanism.
      virtual void HandleEvents() = 0;

      // Overrideable initialization event handlers
      //!
      //! This function can be overriden by user to handle "MainLoopEnter" event.
      virtual bool EventMainLoopEnter();
      //!
      //! This function can be overriden by user to handle "MainLoopLeave" event.
      virtual bool EventMainLoopLeave();

      // Overrideable keyboard event handlers
      //!
      //! This function can be overriden by user to handle "KeyPress" event.
      //! \deprecated Will be removed soon.
      virtual void EventKeyPress(EKeySymbol eKeySym);
      //!
      //! This function can be overriden by user to handle "KeyPress" event.
      virtual void EventKeyPress(EKeySymbol eKeySym, const char *str);
      //!
      //! This function can be overriden by user to handle "KeyRelease" event.
      //! \deprecated Will be removed soon.
      virtual void EventKeyRelease(EKeySymbol eKeySym);
      //!
      //! This function can be overriden by user to handle "KeyRelease" event.
      virtual void EventKeyRelease(EKeySymbol eKeySym, const char *str);

      // Overrideable mouse event handlers
      //!
      //! This function can be overriden by user to handle "ButtonPress" event.
      //! \deprecated Will be removed soon.
      virtual void EventButtonPress(EBtnSymbol eButtonSym);
      //!
      //! This function can be overriden by user to handle "ButtonPress" event.
      virtual void EventButtonPress(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos);
      //!
      //! This function can be overriden by user to handle "ButtonRelease" event.
      //! \deprecated Will be removed soon.
      virtual void EventButtonRelease(EBtnSymbol eButtonSym);
      //!
      //! This function can be overriden by user to handle "ButtonRelease" event.
      virtual void EventButtonRelease(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos);
      //!
      //! This function can be overriden by user to handle "MouseMove" event.
      //! \deprecated Will be removed soon.
      virtual void EventMouseMove(int32_t n32XPos, int32_t n32YPos);
      //!
      //! This function can be overriden by user to handle "MouseMove" event.
      virtual void EventMouseMove(int32_t n32XPos, int32_t n32YPos, bool bButtonState[5]);

      // Overrideable window event handlers
      //!
      //! This function can be overriden by user to handle "UpdateFrame" event.
      virtual void EventUpdateFrame(bool bWindowIsActive);
      //!
      //! This function can be overriden by user to handle "WindowVisibility" event.
      virtual void EventWindowVisibility(bool bWindowIsVisible);
      //!
      //! This function can be overriden by user to handle "WindowActivated" event.
      virtual void EventWindowActivated();
      //!
      //! This function can be overriden by user to handle "WindowDeactivated" event.
      virtual void EventWindowDeactivated();
      //!
      //! This function can be overriden by user to handle "CloseWindow" event.
      virtual void EventCloseWindow();

      // Overrideable other event handlers
      //!
      //! This function can be overriden by user to handle "PasteData" event.
      virtual void EventPasteData(CClipboard::CDT eDataType);

    protected:
      CVideoEnv *_pcsVidEnv; // Hooked video environment
      bool       _bDoExit;   // Exit flag
      bool       _bActive;   // Active flag
      bool       _bVisible;  // Visibility flag
  };


  //!
  //! \brief A standard event handler class
  //!
  //! \api beta
  class VGE_EXPORT CEventHandler : public CEventHandlerBase {
    public:
      //!
      //! Construct a new CEventHandler object using the given video environment.
      explicit CEventHandler(CVideoEnv *pcsVidEnv);

      // Overrideable loop functions
      //!
      //! Call this function to run the application main loop.
      virtual int32_t DoMainLoop();

      // Overrideable event handling functions
      //!
      //! Overriden by CEventHandler to implement the event pumping mechanism.
      //! \n\n
      //! Shouldn't need to call this function directly unles the user has
      //! overriden the DoMainLoop() function.
      virtual bool PumpEvents();
      //!
      //! Overriden by CEventHandler to implement the event handling mechanism.
      //! \n\n
      //! Shouldn't need to call this function directly unles the user has
      //! overriden the DoMainLoop() function.
      virtual void HandleEvents();
  };

  //!
  //! \brief An extended event handler class (with FPS control)
  //!
  //! \api beta
  class VGE_EXPORT CEventHandlerExt : public CEventHandlerBase {
    public:
      //!
      //! Construct a new CEventHandler object using the given video environment.
      explicit CEventHandlerExt(CVideoEnv *pcsVidEnv, float fWantedFPS = 30.0f);

      // FPS functions
      //!
      //! Delegated to the CEventHandlerExt's internal CFPSControl instance.
      inline void SaveStartTicks() { _csFPSCtrl.SaveStartTicks(); }
      //!
      //! Delegated to the CEventHandlerExt's internal CFPSControl instance.
      inline void SaveEndTicks() { _csFPSCtrl.SaveEndTicks(); }
      //!
      //! Delegated to the CEventHandlerExt's internal CFPSControl instance.
      inline float GetAvgFPS() const { return(_csFPSCtrl.GetAvgFPS()); }
      //!
      //! Delegated to the CEventHandlerExt's internal CFPSControl instance.
      inline float GetFPS() const { return(_csFPSCtrl.GetFPS()); }
      //!
      //! Delegated to the CEventHandlerExt's internal CFPSControl instance.
      inline void SetWantedFPS(float fWantedFPS) { _csFPSCtrl.SetWantedFPS(fWantedFPS); }
      //!
      //! Enable the FPS information string updating on each loop.
      inline void EnableFPSInfoString() { _bUpdateFPSInfoString = true; }
      //!
      //! Disable the FPS information string updating on each loop.
      inline void DisableFPSInfoString() { _bUpdateFPSInfoString = false; }
      //!
      //! Enable FPS control of the CEventHandlerExt's internal CFPSControl instance.
      inline void EnableFPSControl() { _bDisableFPSControl = false; }
      //!
      //! Disable FPS control of the CEventHandlerExt's internal CFPSControl instance.
      inline void DisableFPSControl() { _bDisableFPSControl = true; }
      //!
      //! Get the FPS information string, affected by EnableFPSInfoString() and
      //! DisableFPSInfoString().
      inline const char *GetFPSInfoString() const { return(_acFPSBuff); };

      // Overrideable loop functions
      //!
      //! Call this function to run the application main loop.
      virtual int32_t DoMainLoop();

      // Overrideable event handling functions
      //!
      //! Overriden by CEventHandlerExt to implement the event pumping mechanism.
      //! \n\n
      //! Shouldn't need to call this function directly unles the user has
      //! overriden the DoMainLoop() function.
      virtual bool PumpEvents();
      //!
      //! Overriden by CEventHandlerExt to implement the event handling mechanism.
      //! \n\n
      //! Shouldn't need to call this function directly unles the user has
      //! overriden the DoMainLoop() function.
      virtual void HandleEvents();

    protected:
      CFPSControl _csFPSCtrl;            // For frame rate control
      char        _acFPSBuff[128];       // The FPS information string buffer
      bool        _bUpdateFPSInfoString; // Flag to indicate if the FPS information string should be updated
      bool        _bDisableFPSControl;   // Flag to indicate if the FPS control should be disabled
  };

} // namespace vge
#endif
