/** \file FrameRateControl.h */
// ############################################################################################
//  FrameRateControl.h
//  Definition & implementation of CFPSControl class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _HIGH_LEVEL_CLASS___FRAME_RATE_CONTROL_H_
#define _HIGH_LEVEL_CLASS___FRAME_RATE_CONTROL_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"
#include "../System/SystemUtilities.h"


namespace vge {

  // Foward reference from the video system implementation details
  // ("lib/VideoEnvironment/VID_*/*")
  class CVideoEnv;


  //!
  //! \brief An FPS calculator and control class
  //!
  //! Instead of using this class directly use the even handler classes (CEventHandler*).
  //!
  //! \api beta
  class VGE_EXPORT CFPSControl : VCLASS_UNCOPYABLE, VCLASS_UNDERIVABLE {
    public:
      //! Construct a CFPSControl object with the given parameters.
      explicit CFPSControl(float fWantedFPS = 30.0f);

      // FPS control functions
      //!
      //! Pump event and perform delay so that the wanted FPS can be maintained.
      bool PumpEventAndDelay(CVideoEnv &csVidEnv, bool bDisableDelay = false);
      //!
      //! Adjust the delay, must be called on each loop.
      bool AdjustDelay(bool bDisableDelay = false);
      //!
      //! Set the wanted FPS.
      inline void SetWantedFPS(float fWantedFPS) { _fWantedFPS = fWantedFPS; }

      // FPS calculation functions
      //!
      //! Save the start ticks to be used by GetAvgFPS() later.
      inline void SaveStartTicks() { _un64STicks = uint64_t(VGetSystemTicks()); }
      //!
      //! Save the end ticks to be used by GetAvgFPS() later.
      inline void SaveEndTicks(){ _un64ETicks = uint64_t(VGetSystemTicks()); }
      //!
      //! Calculate the average FPS.
      inline float GetAvgFPS() const
      { return(float(double(_un64Frames) / (double(_un64ETicks - _un64STicks) / double(TICKS_PER_SEC)))); }
      //!
      //! Calculate current FPS.
      float CalculateFPS();
      //!
      //! Get the current FPS (without re-calculating it).
      inline float GetFPS() const { return(_fFPS); }

      // Information functions
      //!
      //! Get the contents of the delay counter.
      inline int32_t GetDelayCounter() const { return(_n32Dly); }
      //!
      //! Get the current delay factor.
      inline int32_t GetDelayFactor() const { return(_n32DlyMax); }
      //!
      //! Get the total frames until now.
      inline uint64_t GetTotalFrames() const { return(_un64Frames); }

      // Helper functions
      //!
      //! Returns the current value of the system's time stamp counter.
      inline static uint64_t RDTSC();
      //!
      //! Returns the current value of the system's ticks.
      inline static uint64_t GetTicks()
      { return(uint64_t(VGetSystemTicks())); }
      //!
      //! Convert system tick value to mS.
      inline static float Ticks2mS(uint64_t un64Ticks) 
      { return(float(double(un64Ticks) / double(TICKS_PER_SEC))); }
      //!
      //! Calculate the time interval between the given ticks, the result is in mS.
      inline static float CalcIntervalmS(uint64_t un64StartTicks, uint64_t un64EndTicks)
      { return(float(double(un64EndTicks - un64StartTicks) / double(TICKS_PER_SEC))); }
      //!
      //! Calculate the average FPS from the given total frames and ticks.
      inline static float CalcAvgFPS(uint64_t un64TotFrames, uint64_t un64StartTicks, uint64_t un64EndTicks)
      { return(float(double(un64TotFrames) / (double(un64EndTicks - un64StartTicks) / double(TICKS_PER_SEC)))); }

    protected:
      uint64_t _un64STicks, _un64ETicks; // Start and end ticks
      uint64_t _un64Frames;              // Cumulative frame counter
      int32_t  _n32Frames;               // Frame counter (for calculating FPS)
      int32_t  _n32Dly;                  // Delay counter
      int32_t  _n32DlyMax;               // Maxium delay value
      int32_t  _n32FPSD;                 // Frame rate diferences
      uint64_t _tTicks1;                 // Ticks 1 (for calculating FPS)
      uint64_t _tTicks2;                 // Ticks 2 (for calculating FPS)
      float    _fFPS;                    // Current frame rate
      float    _fWantedFPS;              // Wanted frame rate
  };

  uint64_t CFPSControl::RDTSC()
  {
    uint64_t v;
    asm volatile ("rdtsc" : "=A" (v));
    return(v);
  }

} // namespace vge
#endif
