// ############################################################################################
//  Archiver.cc
//  Implementation of CArchiver and CDeArchiver classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../System/SystemUtilities.h"
#include "Archiver.h"

// C++ headers
#include <fstream>

// 3rd party headers
#include <zlib.h>


//
// Constants used in all archive (*.vaf) files
//
#define _VAF_VER_   100
#define _VAF_MAGIC_ "VAF"


// Chunk size when copying files
#define COPY_DATA_CHUNK_SIZE 8192


namespace vge {

//
// Helper function for reading archives
//
static void _Read_MNum_Ver_UsrData(pistream &is, const char *strmName, vector<char> &retUserData, uint32_t &retNumberOfFiles)
{
  // Read the magic numbers
  const size_t lomn = strlen(_VAF_MAGIC_);
  vector<char> mnb(lomn);
  is.read(&mnb[0], lomn);
  if(memcmp(&mnb[0], _VAF_MAGIC_, lomn))
    VError(VERR_InvalidHeaderData, _NFL_, "Invalid header in archive file \"%s\"", strmName);

  // Read the version
  uint32_t ver;
  is.read(reinterpret_cast<char*>(&ver), sizeof(ver));
  if(ver != _VAF_VER_)
    VError(VERR_IncorrectVersion, _NFL_, "Invalid version number in the header of archive file \"%s\"", strmName);

  // Read the user data
  uint32_t udz;
  is.read(reinterpret_cast<char*>(&udz), sizeof(udz));
  if(udz) {
    retUserData.resize(udz);
    is.read(&retUserData[0], udz);
  }
  else
    retUserData.clear();

  // Read the number of files
  is.read(reinterpret_cast<char*>(&retNumberOfFiles), sizeof(retNumberOfFiles));
}


//
// CArchiver implementation
//
CArchiver::CArchiver(const string &inputFilesBasePath)
: _inputFilesBasePath(inputFilesBasePath)
{
  _FixInputFilesBasePath();
}

CArchiver::CArchiver(const string &inputFilesBasePath, const vector<string> &inputFiles)
: _inputFilesBasePath(inputFilesBasePath), _inputFiles(inputFiles)
{
  _FixInputFilesBasePath();
}

CArchiver::CArchiver(const string &inputFilesBasePath, const vector<string> &inputFiles, const vector<char> &userData)
: _inputFilesBasePath(inputFilesBasePath), _inputFiles(inputFiles), _userData(userData)
{
  _FixInputFilesBasePath();
}

void CArchiver::Clear()
{
  _inputFilesBasePath.clear();
  _inputFiles.clear();
  _userData.clear();
}

void CArchiver::SetUserDataStr(const string &userData)
{
  if(!userData.length()) {
    _userData.clear();
    return;
  }
  _userData.resize(userData.length());
  memcpy(&_userData[0], userData.data(), userData.length());
}

void CArchiver::ProcessFiles(const CDataSink &sink) const
{
  // Get the stream
  const char *strmName = sink.Name().c_str();
  ostream    &os       = sink.Stream();

  // Write the archive
  os.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    // Write the magic numbers
    os.write(_VAF_MAGIC_, strlen(_VAF_MAGIC_));
    // Write the version
    uint32_t ver = _VAF_VER_;
    os.write(reinterpret_cast<char*>(&ver), sizeof(ver));
    // Write the user data
    uint32_t udz = _userData.size();
    os.write(reinterpret_cast<char*>(&udz), sizeof(udz));
    if(udz) os.write(&_userData[0], udz);
    // Write the number of files
    uint32_t nof = _inputFiles.size();
    os.write(reinterpret_cast<char*>(&nof), sizeof(nof));
    // Write the file one by one
    for(size_t i = 0; i < nof; ++i) {
      // Check if the entry is empty
      if(_inputFiles[i].empty()) continue;
      // Get the source file size
      const string srcFile = _inputFilesBasePath + _inputFiles[i];
      TFileInfo ti;
      VGetFileInfo(ti, srcFile);
      // Write the file name
      uint32_t fnz = _inputFiles[i].length();
      os.write(reinterpret_cast<char*>(&fnz), sizeof(fnz));
      os.write(&_inputFiles[i][0], fnz);
      // Write the file size
      uint32_t fz = ti.Size;
      os.write(reinterpret_cast<char*>(&fz), sizeof(fz));
      // Check if the file is empty
      if(!fz) continue;
      // Open the source file
      ifstream fs;
      fs.open(srcFile.c_str());
      if(fs.fail()) VError(VERR_OpenFile, _NFL_, "Couldn't open input file \"%s\" for archiving", srcFile.c_str());
      // Read and write the source file data
      vector<char>   tbuff(COPY_DATA_CHUNK_SIZE);
      const size_t tloop = fz / tbuff.size();
      const size_t trmdr = fz % tbuff.size();
      for(size_t j = 0; j < tloop; ++j) {
        fs.read(&tbuff[0], tbuff.size());
        if(fs.fail()) VError(VERR_ReadStream, _NFL_, "Couldn't read source file data from stream \"file://%s\"", srcFile.c_str());
        os.write(&tbuff[0], tbuff.size());
      }
      if(trmdr) {
        fs.read(&tbuff[0], trmdr);
        if(fs.fail()) VError(VERR_ReadStream, _NFL_, "Couldn't read source file data from stream \"file://%s\"", srcFile.c_str());
        os.write(&tbuff[0], trmdr);
      }
      // Close the source file
      fs.close();
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_WriteStream, _NFL_, "Couldn't write archive data to stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}


//
// CDeArchiver implementation
//
CDeArchiver::CDeArchiver(const string &outputFilesBasePath)
: _outputFilesBasePath(outputFilesBasePath)
{
  _FixOutputFilesBasePath();
}

CDeArchiver::~CDeArchiver() throw()
{
}

void CDeArchiver::Clear()
{
  _outputFilesBasePath.clear();
  _outputFiles.clear();
  _userData.clear();
}

void CDeArchiver::ProcessFiles(const CDataSource &source)
{
  // Get the stream
  const char *strmName = source.Name().c_str();
  pistream   &is       = source.PStream();

  // Read the archive
  is.exceptions(ios::badbit | ios::failbit | ios::eofbit);
  try {
    // Read the magic numbers, version number, user data and number of files
    uint32_t nof;
    _Read_MNum_Ver_UsrData(is, strmName, _userData, nof);
    // Validate the user data
    ValidateUserData(_userData);
    // Read all the files one by one
    _outputFiles.clear();
    for(size_t i = 0; i < nof; ++i) {
      // Read the file name
      uint32_t fnz;
      is.read(reinterpret_cast<char*>(&fnz), sizeof(fnz));
      if(!fnz)
        continue;
      else {
        string rfn;
        rfn.resize(fnz);
        is.read(&rfn[0], fnz);
        _outputFiles.push_back(rfn);
      }
      // Generate the destination file name
      const string dstFile = _outputFilesBasePath + _outputFiles.back();
      // Read the file size
      uint32_t fz;
      is.read(reinterpret_cast<char*>(&fz), sizeof(fz));
      // Check if the file is empty
      if(!fz) {
        VCreateEmptyFile(dstFile);
        continue;
      }
      // Open the destination file
      ofstream fs;
      fs.open(dstFile.c_str());
      if(fs.fail()) VError(VERR_OpenFile, _NFL_, "Couldn't open ouput file \"%s\" for de-archiving", dstFile.c_str());
      // Read and write the source file data
      vector<char>   tbuff(COPY_DATA_CHUNK_SIZE);
      const size_t tloop = fz / tbuff.size();
      const size_t trmdr = fz % tbuff.size();
      for(size_t j = 0; j < tloop; ++j) {
        is.read(&tbuff[0], tbuff.size());
        fs.write(&tbuff[0], tbuff.size());
        if(fs.fail()) VError(VERR_WriteStream, _NFL_, "Couldn't write destination file data to stream \"file://%s\"", dstFile.c_str());
      }
      if(trmdr) {
        is.read(&tbuff[0], trmdr);
        fs.write(&tbuff[0], trmdr);
        if(fs.fail()) VError(VERR_WriteStream, _NFL_, "Couldn't write destination file data to stream \"file://%s\"", dstFile.c_str());
      }
      // Close the destination file
      fs.close();
    }
  }
  catch(ios_base::failure&) {
    VError(VERR_ReadStream, _NFL_, "Couldn't read archive data from stream \"%s\"", strmName);
  }
  catch(...) {
    throw;
  }
}

const string CDeArchiver::GetUserDataStr() const
{
  if(!_userData.size()) return("");

  string ud;
  ud.resize(_userData.size());
  memcpy(&ud[0], &_userData[0], _userData.size());
  return(ud);
}

void CDeArchiver::ValidateUserData(const vector<char>& userData) const
{
  UNUSED_PARAM(userData);
}


//
// CArchiveDataSource implementation
//
CArchiveDataSource::CArchiveDataSource(const string &archiveFileName, istream *userInputStream)
: _ownIS(false), _oex(ios::goodbit), _is(NULL), _ds(NULL)
{
  if(!archiveFileName.empty() || userInputStream) Open(archiveFileName, userInputStream);
}

CArchiveDataSource::~CArchiveDataSource() throw()
{
  Close();
}

void CArchiveDataSource::Close()
{
  delete _ds;

  if(_is) {
    if(_ownIS) {
      _is->exceptions(ios::goodbit);
      delete _is;
    }
    else
      _is->exceptions(_oex);
  }

  _archiveFN.clear();
  _fileList.clear();
  _userData.clear();
  _ownIS = false;
  _oex   = ios::goodbit;
  _is    = NULL;
  _ds    = NULL;
}

void CArchiveDataSource::Open(const string &archiveFileName, istream *userInputStream)
{
  // Close previous archive
  if(_is) Close();

  // Initialize the stream
  if(userInputStream) {
    _archiveFN = archiveFileName;
    _is        = userInputStream;
    _oex       = _is->exceptions();
  }
  else {
    // Create a new ifstream instance
    _is = new ifstream;
    // Open the archive
    reinterpret_cast<ifstream*>(_is)->open(archiveFileName.c_str());
    if(_is->fail()) {
      delete _is;
      _is = NULL;
      VError(VERR_OpenFile, _NFL_, "Couldn't open file \"%s\" for reading", archiveFileName.c_str());
    }
    // Set data
    _archiveFN  = "file://" + archiveFileName;
    _ownIS      = true;
  }
  _is->exceptions(ios::badbit | ios::failbit | ios::eofbit);

  // For convenience
  const char *strmName = _archiveFN.c_str();
  pistream    is(*_is);

  // Read the archive
  try {
    // Read the magic numbers, version number, user data and number of files
    uint32_t nof;
    _Read_MNum_Ver_UsrData(is, strmName, _userData, nof);
    // Read all the files one by one
    _fileList.clear();
    for(size_t i = 0; i < nof; ++i) {
      // Read the file name
      uint32_t fnz;
      is.read(reinterpret_cast<char*>(&fnz), sizeof(fnz));
      if(!fnz) continue;
      string fileName;
      fileName.resize(fnz);
      is.read(&fileName[0], fnz);
      // Read the file size
      uint32_t fileSize;
      is.read(reinterpret_cast<char*>(&fileSize), sizeof(fileSize));
      // Get the start position of the file body (data)
      uint32_t fileStart = is.tellg();
      // Skip the file body (data)
      is.seekg(fileSize, ios_base::cur);
      // Put the information in the list
      _fileList.push_back(TFileListEntry(fileName, fileStart, fileSize));
    }
  }
  catch(ios_base::failure&) {
    Close();
    VError(VERR_ReadStream, _NFL_, "Couldn't read archive data from stream \"%s\"", strmName);
  }
  catch(...) {
    Close();
    throw;
  }

  // Make sure that our istream does not throw by default
  _is->exceptions(ios::goodbit);
}

const string CArchiveDataSource::GetUserDataStr() const
{
  if(!_userData.size()) return("");

  string ud;
  ud.resize(_userData.size());
  memcpy(&ud[0], &_userData[0], _userData.size());
  return(ud);
}

CDataSource &CArchiveDataSource::GetFile(const string &fileName)
{
  for(size_t i = 0; i < _fileList.size(); ++i) {
    if(_fileList[i].name == fileName) {
      // Seek to the start position of the file
      _is->seekg(_fileList[i].start);
      _is->clear();
      // Create a new CDataSource instance
      CDataSource *ds = new CDataSource(*_is, _fileList[i].name.c_str(), _fileList[i].size, _fileList[i].start);
      // Delete the old instance and assign the new instance
      delete _ds; _ds = ds;
      return(*_ds);
    }
  }
  VError(VERR_NotExist, _NFL_, "The requested file \"%s\" is not in the archive \"%s\"",
         fileName.c_str(), _archiveFN.c_str());
}

} // namespace vge
