// ############################################################################################
//  FatalSignalHandler.cc
//  Implementation of fatal signal handler functions
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "FatalSignalHandler.h"

// GCC and Linux specific headers
#include <cxxabi.h>
#include <execinfo.h>


namespace vge {

// From X11 video implementation file,
// "lib/VideoEnvironment/VID_X11/VX11Environment.h"
#ifdef __USE_X11_VIDEO__
extern void _XCrash_RestoreState_();
#endif


// Used by signal handler
static sighandler_t               _atOldHandler_[11] = {SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL, SIG_DFL};
static PFNVortexFatalFuncCallback _pfnUser_          = NULL;
static void                      *_pvUserData_       = NULL;
static bool                       _bInstalled_       = false;

// Handlers
static void _VFSH_(int32_t n32Signal);
static void _AtExit_();


// Install fatal signal handler
void VInstallFSH(PFNVortexFatalFuncCallback pfnUserCallback, void *pvUserData)
{
  // Static variables
  static bool bAtExitInstalled = false;

  // Make sure that this only installed once
  if(_bInstalled_) return;
  _bInstalled_ = true;

  // Install handler
  _pfnUser_ = pfnUserCallback; _pvUserData_ = pvUserData;
  _atOldHandler_[ 0] = signal(SIGINT,  _VFSH_);
  _atOldHandler_[ 1] = signal(SIGQUIT, _VFSH_);
  _atOldHandler_[ 2] = signal(SIGILL,  _VFSH_);
  _atOldHandler_[ 3] = signal(SIGABRT, _VFSH_);
  _atOldHandler_[ 4] = signal(SIGFPE,  _VFSH_);
  _atOldHandler_[ 5] = signal(SIGSEGV, _VFSH_);
  _atOldHandler_[ 6] = signal(SIGPIPE, _VFSH_);
  _atOldHandler_[ 7] = signal(SIGALRM, _VFSH_);
  _atOldHandler_[ 8] = signal(SIGTERM, _VFSH_);
  _atOldHandler_[ 9] = signal(SIGTSTP, _VFSH_);
  _atOldHandler_[10] = signal(SIGBUS,  _VFSH_);

  // Register a function to be called at exit
  if(!bAtExitInstalled) {
    bAtExitInstalled = true;
    atexit(_AtExit_);
  }
}

// Uninstall fatal signal handler
void VUninstallFSH()
{
  // Make sure that this only uninstalled once
  if(!_bInstalled_) return;
  _bInstalled_ = false;

  // Uninstall handler
  _pfnUser_ = NULL; _pvUserData_ = NULL;
  signal(SIGINT,  _atOldHandler_[ 0]); _atOldHandler_[ 0] = SIG_DFL;
  signal(SIGQUIT, _atOldHandler_[ 1]); _atOldHandler_[ 1] = SIG_DFL;
  signal(SIGILL,  _atOldHandler_[ 2]); _atOldHandler_[ 2] = SIG_DFL;
  signal(SIGABRT, _atOldHandler_[ 3]); _atOldHandler_[ 3] = SIG_DFL;
  signal(SIGFPE,  _atOldHandler_[ 4]); _atOldHandler_[ 4] = SIG_DFL;
  signal(SIGSEGV, _atOldHandler_[ 5]); _atOldHandler_[ 5] = SIG_DFL;
  signal(SIGPIPE, _atOldHandler_[ 6]); _atOldHandler_[ 6] = SIG_DFL;
  signal(SIGALRM, _atOldHandler_[ 7]); _atOldHandler_[ 7] = SIG_DFL;
  signal(SIGTERM, _atOldHandler_[ 8]); _atOldHandler_[ 8] = SIG_DFL;
  signal(SIGTSTP, _atOldHandler_[ 9]); _atOldHandler_[ 8] = SIG_DFL;
  signal(SIGBUS,  _atOldHandler_[10]); _atOldHandler_[10] = SIG_DFL;
}

// Block/unblock SIGINT (Ctrl-C)
void VEnableCtrlC(bool bEnabled)
{
  if(bEnabled) {
    if(_bInstalled_) signal(SIGINT, _VFSH_);
    else             signal(SIGINT, SIG_DFL);
  }
  else
    signal(SIGINT, SIG_IGN);
}

// Block/unblock SIGTSTP (Ctrl-Z)
void VEnableCtrlZ(bool bEnabled)
{
  if(bEnabled) {
    if(_bInstalled_) signal(SIGTSTP, _VFSH_);
    else             signal(SIGTSTP, SIG_DFL);
  }
  else
    signal(SIGTSTP, SIG_IGN);
}

// Dump the stack
void VDumpStack()
{
  void  *array[256];
  int    nSize   = backtrace(array, 256);
  char **symbols = backtrace_symbols(array, nSize);

  // Check if backtrace is not available
  if(!nSize) {
    free(symbols);
    VPrint("\nSTACK BACKTRACE: NOT AVAILABLE\n");
    return;
  }

  // Print the title
  VPrint("\nSTACK BACKTRACE:\n");
  VPrint("------------------------------------------------------------------------------------------\n");
  VPrint("Address      Function Name                                                      Location  \n");
  VPrint("------------------------------------------------------------------------------------------\n");

  // Print the stack backtrace
  for (int i = 0; i < nSize; i++) {
    // Copy the values to a local buffer
    char cs[8192];
    safe_strcpy(cs, symbols[i], sizeof(cs));

    // Parse the values
    //
    //               0000000000111111111122222222223333333333
    //               0123456789012345678901234567890123456789
    // ./__app_name__(__function__name__v0+0x41) [0x80488f9]
    //               l                    p    r b         n
    char *l = strchr (cs, '(');
    char *p = strrchr(cs, '+');
    char *r = strrchr(cs, ')');
    char *b = strchr (cs, '[');
    char *n = strrchr(cs, ']');

    // Check if any of the wanted values is not found
    if(!l || !p || !r || !b || !n) {
      if(b && n) {
        *b = 0;
        *n = 0;
        VPrint(b + 1);
      }
      else
        VPrint(symbols[i]);
      VPrint("\n");
      continue;
    }

    // Check if the mangled name is too long
    char   mn[8192];
    size_t sm = p - l - 1;
    if(sm >= sizeof(mn)) {
      VPrint(symbols[i]);
      VPrint("\n");
      continue;
    }

    // If not too long then copy the name
    memset(mn, 0, sizeof(mn));
    strncpy(mn, l + 1, sm);

    // Demangle the name
    char   dn[8192];
    size_t sz = sizeof(dn);
    int    st;
    abi::__cxa_demangle(mn, dn, &sz, &st);

    // Demangle failed
    if(st < 0) {
      *r = 0; *n = 0;
      VPrint("%-11s  %-65s  +%-10s\n", b + 1, mn, p + 1);
    }
    // Demangle succeed
    else {
      char* d = strchr(dn, '(');
      if(d) *d = 0;

      *l = 0; *r = 0; *n = 0;

      VPrint("%-11s  %-65s  +%-10s\n", b + 1, dn, p + 1);
    }
  }

  // Done
  VPrint("------------------------------------------------------------------------------------------\n");
  free(symbols);
}


//
// Local functions implementation
//

// Fatal signal handler so that fatal signals won't cause
// direct termination without any information
static void _VFSH_(int32_t n32Signal)
{
  // Restore the original handler
  VUninstallFSH();

  // Call the user callback functions
  if(_pfnUser_) {
    bool bContinue = _pfnUser_(_pvUserData_) && (n32Signal==SIGINT || n32Signal==SIGTSTP);
    if(bContinue) {
      VInstallFSH();
      return;
    }
  }

  // Block any further signals
  sigset_t tMask;
  sigfillset(&tMask);
  sigprocmask(SIG_SETMASK, &tMask, NULL);

  // Try to restore the system state
#ifdef __USE_X11_VIDEO__
  _XCrash_RestoreState_();
#endif

  // Print message
  try {
    VFatalError(VERR_FATAL_GotFatalSignal, _NFL_, "SIGNAL: %s", strsignal(n32Signal));
  }
  catch(...) {}

  // Dump the stack
  VDumpStack();

  // Unblock the signals before exiting
  sigfillset(&tMask);
  sigprocmask(SIG_UNBLOCK, &tMask, NULL);

  // Exit program
  exit(-n32Signal);
}

// Called at program exit to make sure that the fatal signal handler is uninstalled
static void _AtExit_()
{
  VUninstallFSH();
}

} // namespace vge
