// ############################################################################################
//  Thread.cc
//  Implementation of CThread class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "../System/SystemUtilities.h"
#include "Thread.h"


namespace vge {

//
// The thread class
//
CThread::CThread(PFNVortexThreadCallback pfnUserCallback, void *pvUserData)
: _tThreadHandle(0),     _pfnUserCallback(pfnUserCallback),
  _bExit        (false), _bSuspend       (true),
  _bTEN         (false), _un64ExitStatus (0)
{
  _tTD.pcsThread  = this;
  _tTD.pbExit     = &_bExit;
  _tTD.pbSuspend  = &_bSuspend;
  _tTD.pvUserData = pvUserData;
}

CThread::~CThread() throw()
{
  try        { Stop(); }
  catch(...) { Kill(); }
}

void CThread::Initialize(PFNVortexThreadCallback pfnUserCallback, void *pvUserData)
{
  // Check
  if(_tThreadHandle)
    VError(VERR_InUse, _NFL_,
           "Can't reinitialize the CThread class instance when its\nthread procedure (function) is running");

  // Set variables
  _tThreadHandle   = 0;
  _pfnUserCallback = pfnUserCallback;
  _bExit           = false;
  _bSuspend        = true;
  _bTEN            = false;
  _un64ExitStatus  = 0;
  _tTD.pcsThread   = this;
  _tTD.pbExit      = &_bExit;
  _tTD.pbSuspend   = &_bSuspend;
  _tTD.pvUserData  = pvUserData;
}

void CThread::Start()
{
  // Temporary variables
  pthread_attr_t attr;

  // Initialize attributes
  if(pthread_attr_init(&attr)) {
    VError(VERR_NotInitialized, _NFL_, "Couldn't initialize the thread attributes");
  }
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

  // Create thread
  if(pthread_create(&_tThreadHandle, &attr, CThread::_ThreadProc, &_tTD))
    VError(VERR_CreateThread, _NFL_, "Couldn't create thread");

  // Free the attributes structure
  pthread_attr_destroy(&attr);

  // Set variables
  _bSuspend = false;
  _bExit    = false;
  _bTEN     = false;
}

void CThread::Stop(uint32_t un32TimeOut_mS)
{
  // Check
  if(_tThreadHandle == 0) return;;

  // Order the thread to exit itself and wait
  uint64_t tC1, tC2;
  tC1 = tC2 = VGetSystemTicks();
  _bExit = true;
  while(uint64_t((tC2 - tC1) / TICKS_PER_SEC * 1000) < un32TimeOut_mS) {
    _bExit = true;
    tC2 = VGetSystemTicks();
    if(_bTEN) {
      _tThreadHandle = 0;
      _bExit         = false;
      _bTEN          = false;
      return;
    }
  }

  // Failed !
  VError(VERR_NotResponding, _NFL_, "The thread is not responding to the Stop() command");
}

void CThread::Kill()
{
  // Check
  if(_tThreadHandle == 0) return;

  // Force stop thread
  if(pthread_cancel(_tThreadHandle)) {
    VWarning(_NFL_, "Thread termination using pthread_cancel() failed; trying pthread_kill()");
    if(pthread_kill(_tThreadHandle, SIGKILL)) {
      VWarning(_NFL_, "Thread termination using pthread_kill() also failed; the thread is still running");
      return;
    }
  }
  VInfo(_NFL_, "The thread is forcedly killed");

  // Set variables
  _tThreadHandle = 0;
  _bTEN          = false;
}

uint64_t CThread::ThreadCallback(TThreadData &threadData)
{
  UNUSED_PARAM(threadData);
  return(0);
}

void *CThread::_ThreadProc(void *pvData)
{
  static const int32_t _an32TMSig_[] = { SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGALRM, SIGTERM,
                                         SIGCHLD, SIGWINCH, SIGVTALRM, SIGPROF };

  // Mask asynchronous signals for this thread
  sigset_t tSig;
  sigemptyset(&tSig);
  for(size_t c = 0; c < 10; ++c)
    sigaddset(&tSig, _an32TMSig_[c]);
  pthread_sigmask(SIG_BLOCK, &tSig, 0);

  // Allow this thread to be asynchronously canceled
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  // Execute the user callback function
  PTThreadData td = reinterpret_cast<PTThreadData>(pvData);

  if(td->pcsThread->_pfnUserCallback)
    td->pcsThread->_un64ExitStatus = td->pcsThread->_pfnUserCallback(td);
  else
    td->pcsThread->_un64ExitStatus = td->pcsThread->ThreadCallback(*td);
  td->pcsThread->_bTEN = true;

  // End thread
  pthread_exit(NULL);
}


//
// The mutex class
//
CMutex::CMutex()
: _tOwner(0)
{
  // Temporary variables
  pthread_mutexattr_t tAttr;

  // Initialize mutex
  pthread_mutexattr_init(&tAttr);
  pthread_mutex_init(&_tMutex, &tAttr);
}

CMutex::~CMutex() throw()
{
  if(_tOwner) pthread_mutex_unlock(&_tMutex); // Might be useless
  pthread_mutex_destroy(&_tMutex);
}

void CMutex::Lock()
{
  // Check
  if(_tOwner == pthread_self()) return;

  // Lock mutex
  _if_likely(pthread_mutex_lock(&_tMutex) == 0) {
    _tOwner = pthread_self();
    return;
  }
  VError(VERR_LockMutex, _NFL_, "Couldn't lock mutex");
}

void CMutex::Unlock()
{
  // Check
  if(_tOwner == 0) return;
  if(_tOwner != pthread_self())
    VError(VERR_MutexNotOwned, _NFL_, "Can't unlock a not owned mutex");

  // Unlock mutex
  _if_likely(pthread_mutex_unlock(&_tMutex) == 0) {
    _tOwner = 0;
    return;
  }
  VError(VERR_UnlockMutex, _NFL_, "Failed to unlock mutex");
}

void CMutex::TryLock(uint32_t un32TimeOut_uS)
{
  // Check
  if(_tOwner == pthread_self()) return;

  // Preparation
  struct timeval tCur, tOut;
  gettimeofday(&tCur, 0);
  tOut.tv_sec  = tCur.tv_sec + (un32TimeOut_uS / 1000);
  tOut.tv_usec = (tCur.tv_usec + (un32TimeOut_uS % 1000) * 1000);

  // Try to lock mutex
  for(;;) {
    // Try lock
    int32_t n32Ret = pthread_mutex_trylock(&_tMutex);
    if(n32Ret == 0) {
      _tOwner = pthread_self();
      return;
    }
    if(n32Ret != EBUSY)
      VError(VERR_LockMutex, _NFL_, "Failed to lock mutex");
    // Check for timeout
    if(!un32TimeOut_uS) break;
    gettimeofday(&tCur, 0);
    if(tCur.tv_sec>tOut.tv_sec || (tCur.tv_sec==tOut.tv_sec && tCur.tv_usec>tOut.tv_usec))
      break;
  }

  // .... Timedout
  VError(VERR_Timeout, _NFL_, "Timedout when trying to lock mutex");
}

} // namespace vge
