/** \file DialogLayout.h */
// ############################################################################################
//  DialogLayout.h
//  Definition & implementation of VortexGE layout controls
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VIDEO_ENVIRONMENT___DIALOG_LAYOUT_H_
#define _VIDEO_ENVIRONMENT___DIALOG_LAYOUT_H_

// VortexGE headers
#include "DialogBase.h"


namespace vge {

  //!
  //! \brief Generic-layout control
  //!
  //! \api beta
  class VGE_EXPORT CGLayout : public _CControl {
    public:
      //!
      //! \brief Helper structure used to store the layout classes's blocks
      //!
      //! \api beta
      struct TBlock {
        int32_t  x1, y1, x2, y2;
        uint32_t w, h;

        inline TBlock(int32_t x1_ = 0, int32_t y1_ = 0, int32_t x2_ = 0, int32_t y2_ = 0, uint32_t w_ = 0, uint32_t h_ = 0)
        : x1(x1_), y1(y1_), x2(x2_), y2(y2_), w(w_), h(h_)
        {}
      };

      //! Default pixel padding size (for border and spacing)
      static const uint32_t DEFAULT_PAD_SIZE = 5;

      //! Constant to be used for getting the whole control's area
      //! as a TBlock instance.
      static const uint32_t WHOLE_AREA = 0xFFFFFFFF;

    public:
      virtual ~CGLayout() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv           The video environment in which this control will be used.
      //! \param[in] parent           Parent control of this control.
      //! \param[in] parentBlockIndex Parent's block to be occupied (used only if the parent control is another layout class).
      //! \param[in] borderWidth      Border size (in pixels).
      //! \param[in] spacing          Spacing between blocks (in pixels).
      void InitControl(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex = 0, uint32_t borderWidth = DEFAULT_PAD_SIZE, uint32_t spacing = DEFAULT_PAD_SIZE);
      //!
      //! Create a new block (for layouting).
      //! \param[in] size Size of the block, zero means to divide equally
      //!                 amongs all the blocks.
      //!
      //! Block is constructed from top to bottom or from left to right, it means
      //! that the 1st block added will be the top-most or left-most block, and the
      //! last block added will be the bottom-most or right-most block.
      inline void AddBlock(uint32_t size = 0)
      { _blocks.push_back(size); _bdirty = true; }
      //!
      //! Create new blocks (for layouting).
      //! \param[in] count Number of blocks to be added.
      //! \param[in] size  Size of the block, zero means to divide equally
      //!                  amongs all the blocks.
      //!
      //! Block is constructed from top to bottom or from left to right, it means
      //! that the 1st block added will be the top-most or left-most block, and the
      //! last block added will be the bottom-most or right-most block.
      void AddBlocks(size_t count, uint32_t size = 0);
      //!
      //! Create new blocks (for layouting).
      //! \param[in] count Number of blocks to be added.
      //! \param[in] size  Size of the block(s), zero means to divide equally
      //!                  amongs all the blocks. If the number of the actual
      //!                  parameters given is less than the number specified
      //!                  in the "count" parameter, the result is undefined.
      //!
      //! Block is constructed from top to bottom or from left to right, it means
      //! that the 1st block added will be the top-most or left-most block, and the
      //! last block added will be the bottom-most or right-most block.
      void AddBlocksEx(size_t count, uint32_t size, ...);

      // Information functions
      //!
      //! Returns the number of blocks
      inline size_t GetBlockCount() const
      { return(_blocks.size()); }

      // Access to the blocks (for layouting)
      //!
      //! Get read-only access to the blocks.
      //! \param[in] index Index to a particular layout block, CGLayout::WHOLE_AREA
      //!                  will return the whole client's area of the control.
      const TBlock &operator[](uint32_t index) const;
  
      // Drawing functions, implemented as dummy since layout controls
      // can't be drawn.
      virtual void DrawControl() const
      {}
      virtual void DrawControl(int32_t mouseX, int32_t mouseY, bool pressed = false) const
      { UNUSED_PARAMS_3(mouseX, mouseY, pressed); }
  
    protected:
      // Helper functions
      virtual ECtrlID _CtrlID()          const = 0;
      virtual void    _CalcBlocksAreas() const = 0;
  
#define _vmcm mutable
      // Data
      uint32_t         _spacing;     // Spacing
      uint32_t         _borderWidth; // Border width
      TBlock           _marea;       // Main area
_vmcm bool             _bdirty;      // Flags to indicate that the blocks' areas need to be recalculated
      vector<uint32_t> _blocks;      // Array of blocks
_vmcm vector<TBlock>   _bareas;      // Array of blocks' areas
#undef _vmcm
  };


  //!
  //! \brief Horizontal-layout control
  //!
  //! \api beta
  class VGE_EXPORT CHLayout : public CGLayout {
    public:
      virtual ~CHLayout() throw();

    protected:
      // Helper functions
      virtual ECtrlID _CtrlID()       const { return(ECID_HLayout); }
      virtual void _CalcBlocksAreas() const;
  };


  //!
  //! \brief Vertical-layout control
  //!
  //! \api beta
  class VGE_EXPORT CVLayout : public CGLayout {
    public:
      virtual ~CVLayout() throw();
  
    protected:
      // Helper functions
      virtual ECtrlID _CtrlID()       const { return(ECID_VLayout); }
      virtual void _CalcBlocksAreas() const;
  };


  //!
  //! \brief HV-layout-compound class (not a control class)
  //!
  //! \api beta
  class VGE_EXPORT CHVLayoutCompound : VCLASS_UNCOPYABLE {
    public:
      virtual ~CHVLayoutCompound() throw();

      // Initialization functions
      //!
      //! Initialize the compound using the given parameters.
      //! \param[in] vidEnv           The video environment in which this control will be used.
      //! \param[in] parent           Parent control of this compound.
      //! \param[in] parentBlockIndex Parent's block to be occupied (used only if the parent control is another layout class).
      //! \param[in] borderWidth      Border size (in pixels).
      //! \param[in] spacing          Spacing between blocks (in pixels).
      void InitCompound(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex = 0, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE);
      //!
      //! Create a new row (for layouting).
      //! \param[in] rowSize Size of the row, zero means to divide equally
      //!                    amongs all the rows.
      //!
      //! Row is constructed from top to bottom, which means that the 1st row added will
      //! be the top-most row, and the last row added will be the bottom-most row.
      void AddRow(uint32_t rowSize = 0);
      //!
      //! Create new rows (for layouting).
      //! \param[in] rowCount Number of rows to be added.
      //! \param[in] rowSize  Size of the rows, zero means to divide equally
      //!                     amongs all the rows.
      //!
      //! Row is constructed from top to bottom, which means that the 1st row added will
      //! be the top-most row, and the last row added will be the bottom-most row.
      void AddRows(size_t rowCount, uint32_t rowSize = 0);
      //!
      //! Create new rows (for layouting).
      //! \param[in] rowCount Number of rows to be added.
      //! \param[in] rowSize  Size of the row(s), zero means to divide equally
      //!                     amongs all the rows. If the number of the actual
      //!                     parameters given is less than the number specified
      //!                     in the "rowCount" parameter, the result is undefined.
      //!
      //! Row is constructed from top to bottom, which means that the 1st row added will
      //! be the top-most row, and the last row added will be the bottom-most row.
      void AddRowsEx(size_t rowCount, uint32_t rowSize, ...);

      // Access to the columns (vertical layouts)
      //!
      //! Direct access to vertical layout (CVLayout).
      //! \param[in] index Index to a particular row.
      inline CVLayout &operator[](size_t rowIndex)
      { return(_GetVLayout(rowIndex)); }
      //!
      //! Direct access to vertical layout (CVLayout).
      //! \param[in] index Index to a particular row.
      inline const CVLayout &operator[](size_t rowIndex) const
      { return(_GetVLayout(rowIndex)); }

    protected:
      // Helper functions
      void _DeleteVLayouts();
      CVLayout &_GetVLayout(size_t rowIndex) const;

#define _vmcm mutable
      // Data
      CHLayout           _hls;     // Main horizontal layout
_vmcm vector<CVLayout*>  _vls;     // Vertical layouts
      CVideoEnv         *_vidEnv;  // CVideoEnv instance that used to initialize this compound
      uint32_t           _spacing; // Spacing
#undef _vmcm
  };


  //!
  //! \brief VH-layout-compound class (not a control class)
  //!
  //! \api beta
  class VGE_EXPORT CVHLayoutCompound : VCLASS_UNCOPYABLE {
    public:
      virtual ~CVHLayoutCompound() throw();

      // Initialization functions
      //!
      //! Initialize the compound using the given parameters.
      //! \param[in] vidEnv           The video environment in which this control will be used.
      //! \param[in] parent           Parent control of this compound.
      //! \param[in] parentBlockIndex Parent's block to be occupied (used only if the parent control is another layout class).
      //! \param[in] borderWidth      Border size (in pixels).
      //! \param[in] spacing          Spacing between blocks (in pixels).
      void InitCompound(CVideoEnv &vidEnv, const _CControl *parent, size_t parentBlockIndex = 0, uint32_t borderWidth = CGLayout::DEFAULT_PAD_SIZE, uint32_t spacing = CGLayout::DEFAULT_PAD_SIZE);
      //!
      //! Create a new column (for layouting).
      //! \param[in] colSize Size of the column, zero means to divide equally
      //!                    amongs all the columns.
      //!
      //! Column is constructed from left to right, which means that the 1st
      //! column added will be the left-most column, and the last column added
      //! will be the right-most column.
      void AddCol(uint32_t colSize = 0);
      //!
      //! Create new columns (for layouting).
      //! \param[in] colCount Number of columns to be added.
      //! \param[in] colSize  Size of the columns, zero means to divide equally
      //!                     amongs all the columns.
      //!
      //! Column is constructed from left to right, which means that the 1st
      //! column added will be the left-most column, and the last column added
      //! will be the right-most column.
      void AddCols(size_t colCount, uint32_t colSize = 0);
      //!
      //! Create new columns (for layouting).
      //! \param[in] colCount Number of columns to be added.
      //! \param[in] colSize  Size of the column(s), zero means to divide equally
      //!                     amongs all the columns. If the number of the actual
      //!                     parameters given is less than the number specified
      //!                     in the "colCount" parameter, the result is undefined.
      //!
      //! Column is constructed from left to right, which means that the 1st
      //! column added will be the left-most column, and the last column added
      //! will be the right-most column.
      void AddColsEx(size_t colCount, uint32_t colSize, ...);

      // Access to the rows (horizontal layouts)
      //!
      //! Direct access to horizontal layout (CHLayout).
      //! \param[in] index Index to a particular column.
      inline CHLayout &operator[](size_t colIndex)
      { return(_GetHLayout(colIndex)); }
      //!
      //! Direct access to horizontal layout (CHLayout).
      //! \param[in] index Index to a particular column.
      inline const CHLayout &operator[](size_t colIndex) const
      { return(_GetHLayout(colIndex)); }

    protected:
      // Helper functions
      void _DeleteHLayouts();
      CHLayout &_GetHLayout(size_t colIndex) const;

#define _vmcm mutable
      // Data
      CVLayout           _vls;     // Main vertical layout
_vmcm vector<CHLayout*>  _hls;     // Horizontal layouts
      CVideoEnv         *_vidEnv;  // CVideoEnv instance that used to initialize this compound
      uint32_t           _spacing; // Spacing
#undef _vmcm
  };

} // namespace vge
#endif
