// ############################################################################################
//  VX11Environment1.cc
//  Implementation of CClipboard class and some helper functions for CVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VideoEnvironment.cc"


// VortexGE headers
#include "../../HighLevelClass/ApplicationObject.h"
#include "../CommonDialogPages.h"
#include "VX11Cursor.h"
#include "VX11Icon.h"
#include "VX11Utilities.h"


namespace vge {

// Helper macros for handling error on XShmAttach()
#define _TRAP_XSHM_ATTACH_ERROR_BEGIN_()               \
  _pfnXOldHandler_ = XSetErrorHandler(_XShmAttachEH_); \
  _bXShmError_     = false;

#define _TRAP_XSHM_ATTACH_ERROR_END_(tDS)       \
  XSetErrorHandler(_pfnXOldHandler_);           \
  if(_bXShmError_)                              \
  {                                             \
    shmdt(tDS.tShmInfo.shmaddr);                \
    shmctl(tDS.tShmInfo.shmid, IPC_RMID, NULL); \
    tDS.tShmInfo.shmaddr = (char *) -1;         \
    tDS.tShmInfo.shmid   = -1;                  \
    VError(VERR_XShmAttach, _NFL_,              \
           "XShmAttach() : %s",                 \
           _acXShmErrorMessage_);               \
  }


// Special data for fullscreen handling
static CVideoEnv              *_pcsFSOwner_ = NULL; // Owner of the fullscreen window
static CVideoEnv              *_pcsFSHoldr_ = NULL; // Holder of the fullscreen window (when the window is not active)
static __intexp_XF86SavedMode_ _tSavedVideoMode_;   // Saved video mode


//
// For handling error on XShmAttach()
//
#ifdef __USE_XSHM__
static int32_t (*_pfnXOldHandler_)(Display *, XErrorEvent *);
static bool _bXShmError_;
static char _acXShmErrorMessage_[1024];

static int32_t _XShmAttachEH_(Display *ptDpy, XErrorEvent *ptEV)
{
  _bXShmError_ = true;
  XGetErrorText(ptDpy, ptEV->error_code, _acXShmErrorMessage_, sizeof(_acXShmErrorMessage_) - 1);
  return(0);
}
#endif


//
// For restoring X Window state on application crash (well, hopefully)
//
void _XCrash_RestoreState_()
{
  // Open a new temporary X connection since on application
  // crash any existing Display* may become invalid
  Display *ptTmpDisplay = XOpenDisplay(NULL);

  // If it is failed too, oh well :(
  if(!ptTmpDisplay) return;

  // Turn on auto repeat
  XAutoRepeatOn(ptTmpDisplay);
  
  // Restore the video mode. Note that if the fullscreen state variables
  // got corrupted in the crash, this most likely won't restore the original
  // video mode, nothing we can do about this too :(
  if((_pcsFSOwner_ != NULL) && (_pcsFSHoldr_ == NULL)) {
    __intexp_XF86RestoreMode_(ptTmpDisplay, XDefaultScreen(ptTmpDisplay), &_tSavedVideoMode_);
    _pcsFSOwner_ = NULL;
    _pcsFSHoldr_ = NULL;
  }

  // Close the temporary display
  XCloseDisplay(ptTmpDisplay);
}


//
// For handling X fatal I/O error
//
static int32_t _XFatalIO_(Display *) _attr_noreturn;
static int32_t _XFatalIO_(Display *)
{
  _XCrash_RestoreState_();
  VAbortOnFatalError(true);
  VFatalError(VERR_FATAL_IOError, _NFL_, "Fatal X I/O error (connection with the X server is lost)");
}


//
// Video storage class implementation (X11)
//
CVideoStorage::CVideoStorage()
#ifdef __USE_XSHM__
  : _pcsVEnv(NULL), _un8SCnt(0), _ppucPS(NULL), _un32ShmS(0)
#else
  : _pcsVEnv(NULL), _un8SCnt(0), _ptPS(NULL)
#endif
{
}

CVideoStorage::CVideoStorage(CVideoEnv &vidEnv, uint8_t un8SlotCount)
#ifdef __USE_XSHM__
  : _pcsVEnv(NULL), _un8SCnt(0), _ppucPS(NULL), _un32ShmS(0)
#else
  : _pcsVEnv(NULL), _un8SCnt(0), _ptPS(NULL)
#endif
{
  Initialize(vidEnv, un8SlotCount);
}

CVideoStorage::~CVideoStorage() throw()
{ 
  FreeStorage(); 
}

void CVideoStorage::Initialize(CVideoEnv &vidEnv, uint8_t un8SlotCount)
{
  // Check
  vassert(un8SlotCount>=1 && un8SlotCount<=_MAX_STGSLOT_,
          "The number of storage slot must between 1 and %d", _MAX_STGSLOT_);

  // Allocate memory
#ifdef __USE_XSHM__
  typedef u_char *_vpu_char;
  u_char **ppucPS = _nothrow_new _vpu_char[un8SlotCount];
  if(ppucPS == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to allocate video storage");
#else
  Pixmap *ptPS = _nothrow_new Pixmap[un8SlotCount];
  if(ptPS == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to allocate video storage");
#endif

  // Create pixmaps
#ifdef __USE_XSHM__
  uint32_t un32Sz;
  if(vidEnv._tX11.un32Depth == 8)
    un32Sz = vidEnv._tX11.un32Width * vidEnv._tX11.un32Height;
  else if(vidEnv._tX11.un32Depth == 16)
    un32Sz = vidEnv._tX11.un32Width * vidEnv._tX11.un32Height * 2;
  else
    un32Sz = vidEnv._tX11.un32Width * vidEnv._tX11.un32Height * 4;
  for(size_t c = 0; c < un8SlotCount; c++)
    ppucPS[c] = reinterpret_cast<u_char*>(malloc(un32Sz));
#else
  for(size_t c = 0; c < un8SlotCount; c++)
    ptPS[c] = XCreatePixmap(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainWindow,
      vidEnv._tX11.un32Width, vidEnv._tX11.un32Height,
      vidEnv._tX11.un32Depth);
#endif

  // Free previous storage, if any
  FreeStorage();

  // Copy variables
#ifdef __USE_XSHM__
  _ppucPS   = ppucPS;
  _un32ShmS = un32Sz;
#else
  _ptPS     = ptPS;
#endif
  _pcsVEnv  = &vidEnv;
  _un8SCnt  = un8SlotCount;
}

void CVideoStorage::FreeStorage()
{
  // Free pixmaps
#ifdef __USE_XSHM__
  if(_ppucPS) {
    for(size_t c = 0; c < _un8SCnt; ++c)
      if(_ppucPS[c]) free(_ppucPS[c]);
    delete[] _ppucPS;
  }
  _ppucPS   = NULL;
  _un32ShmS = 0;
#else
  if(_ptPS) {
    for(size_t c = 0; c < _un8SCnt; ++c)
      if(_ptPS[c]) XFreePixmap(_pcsVEnv->_tX11.ptDisplay, _ptPS[c]);
    delete[] _ptPS;
  }
  _ptPS = NULL;
#endif
  _pcsVEnv = NULL;
  _un8SCnt = 0;
}

bool CVideoStorage::StoreCurrentView(uint8_t un8DstSlotIndex)
{
  _if_unlikely(un8DstSlotIndex >= _un8SCnt) return(false);
#ifdef __USE_XSHM__
  XSync(_pcsVEnv->_tX11.ptDisplay, False);
  memcpy(_ppucPS[un8DstSlotIndex], _pcsVEnv->_tX11.tShmInfo.shmaddr, _un32ShmS);
#else
  XCopyArea(_pcsVEnv->_tX11.ptDisplay, _pcsVEnv->_tX11.tMainPixmap,
            _ptPS[un8DstSlotIndex], _pcsVEnv->_tX11.tMainGC,
            0, 0, _pcsVEnv->_tX11.un32Width, _pcsVEnv->_tX11.un32Height, 0, 0);
#endif
  return(true);
}

bool CVideoStorage::RestoreCurrentView(uint8_t un8SrcSlotIndex)
{
  _if_unlikely(un8SrcSlotIndex >= _un8SCnt) return(false);
#ifdef __USE_XSHM__
  memcpy(_pcsVEnv->_tX11.tShmInfo.shmaddr, _ppucPS[un8SrcSlotIndex], _un32ShmS);
  XSync(_pcsVEnv->_tX11.ptDisplay, False);
#else
  XCopyArea(_pcsVEnv->_tX11.ptDisplay, _ptPS[un8SrcSlotIndex],
            _pcsVEnv->_tX11.tMainPixmap, _pcsVEnv->_tX11.tMainGC,
            0, 0, _pcsVEnv->_tX11.un32Width, _pcsVEnv->_tX11.un32Height, 0, 0);
#endif
  _pcsVEnv->UpdateDisplay();
  return(true);
}


//
// Clipboard class implemenation (X11)
//
CVideoEnv *CClipboard::_vidEnv = NULL;

void CClipboard::CopyTo(const string &srcString, bool clipboard)
{
  if(!_vidEnv) VError(VERR_NotInitialized, _NFL_, "No CVideoEnv class is associated with this CClipboard class instance");
  __intexp_X11CopyStringToClipboard(_vidEnv->_tX11.ptDisplay, _vidEnv->_tX11.tMainWindow, srcString, clipboard);
}

void CClipboard::RequestData(CDT dataType, bool clipboard)
{
  if(!_vidEnv) VError(VERR_NotInitialized, _NFL_, "No CVideoEnv class is associated with this CClipboard class instance");
  switch(dataType) {
    case CDT_Null:
      break;
    case CDT_String:
      __intexp_X11CopyStringFromClipboard_Request(_vidEnv->_tX11.ptDisplay, _vidEnv->_tX11.tMainWindow, clipboard);
      break;
    default:
      VError(VERR_InvalidParameter, _NFL_, "Requesting for invalid clipboard data type");
  }
}

void CClipboard::CopyFrom(string &dstString)
{
  if(!_vidEnv) VError(VERR_NotInitialized, _NFL_, "No CVideoEnv class is associated with this CClipboard class instance");
  __intexp_X11CopyStringFromClipboard_Acquire(dstString);
}

} // namespace vge
