// ############################################################################################
//  VX11Environment2.cc
//  Implementation of CVideoEnv class using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VideoEnvironment.cc"


namespace vge {

// For experiment, just leave it as it
#define __USE_XIM__
#define __USE_XOM__


//
// Video enviroment class implementation (X11)
//
EKeySymbol CVideoEnv::_VKS[_MAX_X11_VKS_];
EBtnSymbol CVideoEnv::_VBS[_MAX_X11_VBS_];

CVideoEnv::CVideoEnv()
: _bAllInitialized(false), _bActive(false), _bVisible(false),
  _hie_bFocus(true),       _hie_bMap(true), _hie_bVisible(true),
  _ConvertRB2RI_(NULL)
{
  // Set this video environment to the cliboard class
  if(!CClipboard::GetVideoEnv()) CClipboard::SetVideoEnv(this);

  memset(&_tX11,      0, sizeof(_tX11));
  memset(&_tX11CS,    0, sizeof(_tX11CS));
  memset(&_tEB,       0, sizeof(_tEB));
  memset(&_tRI,       0, sizeof(_tRI));
  memset(&_tRB,       0, sizeof(_tRB));
  memset(&_tX11Style, 0, sizeof(_tX11Style));

#ifdef __USE_XSHM__
  _tRI.tShmInfo.shmid  = -1;
  _tX11.tShmInfo.shmid = -1;
#endif
  _tEB.ptEvent = &_tEB.tEmptyEvent;

  CVideoEnv::_InitX11Symbols_();
}

CVideoEnv::~CVideoEnv() throw()
{
  // Remove this video environment from the cliboard class
  if(CClipboard::GetVideoEnv() == this) CClipboard::SetVideoEnv(NULL);

  // Restore the video mode (if we own the fullscreen window)
  if(_pcsFSOwner_ == this) {
    __intexp_XF86RestoreMode_(_tX11.ptDisplay, _tX11.n32Screen, &_tSavedVideoMode_);
    _pcsFSOwner_ = NULL;
  }

  // Free the render buffer datas
  if(_tRB.pn32ZBuffer) abmfree(_tRB.pn32ZBuffer);
  if(_tRB.pun8Buffer)  abmfree(_tRB.pun8Buffer);

  // Free the render image datas
  if(_tRI.pucLUT)   free(_tRI.pucLUT);
  if(_tRI.pun16Pal) free(_tRI.pun16Pal);
#ifdef __USE_XSHM__
  if(_tRI.ptImage) {
    _tRI.ptImage->data = 0;
    XDestroyImage(_tRI.ptImage);
  }
#elif defined(_ALIGN_BUFF_MEM_)
  if(_tRI.ptImage) {
    // Make sure that the X frees the original address
    _tRI.ptImage->data = reinterpret_cast<char*>(ssegetrealaddress(_tRI.pucImage));
    XDestroyImage(_tRI.ptImage);
  }
  else if(_tRI.pucImage)
    ssefree(_tRI.pucImage);
#else
  if(_tRI.ptImage)
    XDestroyImage(_tRI.ptImage);
  else if(_tRI.pucImage)
    free(_tRI.pucImage);
#endif

  // Free the main X11 objects
#ifdef __USE_XOM__
  if(_tX11.tMainXOM) XCloseOM(_tX11.tMainXOM);
#endif
#ifdef __USE_XIM__
  if(_tX11.tMainXIC) XDestroyIC(_tX11.tMainXIC);
  if(_tX11.tMainXIM) XCloseIM(_tX11.tMainXIM);
#endif

  if(_tX11.tMainWindow)    XUndefineCursor(_tX11.ptDisplay, _tX11.tMainWindow);
  if(_tX11.tDefCurArrow)   XFreeCursor(_tX11.ptDisplay, _tX11.tDefCurArrow);
  if(_tX11.tDefCurWait)    XFreeCursor(_tX11.ptDisplay, _tX11.tDefCurWait);
  if(_tX11.tMainPixmap)    XFreePixmap(_tX11.ptDisplay, _tX11.tMainPixmap);
  if(_tX11.tMainIcon)      XFreePixmap(_tX11.ptDisplay, _tX11.tMainIcon);
  if(_tX11.tMainIconMask)  XFreePixmap(_tX11.ptDisplay, _tX11.tMainIconMask);
  if(_tX11.tMainWindow) {
    XUnmapWindow(_tX11.ptDisplay, _tX11.tMainWindow);
    XDestroyWindow(_tX11.ptDisplay, _tX11.tMainWindow);
  }

  if(_tX11.ptDefFontSetSmall)      FreeUtf8Font(_tX11.ptDefFontSetSmall);
  if(_tX11.ptDefFontSetBig)        FreeUtf8Font(_tX11.ptDefFontSetBig);
  if(_tX11.ptDefFontBig)           FreeFont(_tX11.ptDefFontBig);
  if(_tX11.ptDefFontSmall)         FreeFont(_tX11.ptDefFontSmall);
  if(_tX11.tMainGC)                XFreeGC(_tX11.ptDisplay, _tX11.tMainGC);
  if(_tX11.tColormapData.count)    XFreeColormap(_tX11.ptDisplay, _tX11.tColormap);
  if(_tX11.tColormapData.ptColors) free(_tX11.tColormapData.ptColors);

  // Free the shared memory segments used by XShm
#ifdef __USE_XSHM__
  if(_tRI.tShmInfo.shmid != -1) {
    if(_tRI.tShmInfo.shmaddr != reinterpret_cast<char*>(-1)) {
      XShmDetach(_tX11.ptDisplay, &_tRI.tShmInfo);
      XSync(_tX11.ptDisplay, True);
      shmdt(_tRI.tShmInfo.shmaddr);
    }
    shmctl(_tRI.tShmInfo.shmid, IPC_RMID, NULL);
  }
  if(_tX11.tShmInfo.shmid != -1) {
    if(_tX11.tShmInfo.shmaddr != reinterpret_cast<char*>(-1)) {
      XShmDetach(_tX11.ptDisplay, &_tX11.tShmInfo);
      XSync(_tX11.ptDisplay, True);
      shmdt(_tX11.tShmInfo.shmaddr);
    }
    shmctl(_tX11.tShmInfo.shmid, IPC_RMID, NULL);
  }
#endif

  // Close connection
  if(_tX11.ptDisplay) {
    XAutoRepeatOn(_tX11.ptDisplay);
    XCloseDisplay(_tX11.ptDisplay);
  }
}

void CVideoEnv::DoQuickInit(const char *utf8WindowTitle,
                            uint32_t un32WindowWidth, uint32_t un32WindowHeight,
                            uint32_t un32RBuffWidth, uint32_t un32RBuffHeight)
{
  // Get the user preference on color palette
  bool uecm = false;
  char *env = getenv("VGE_VIDEO_CUSTOM_8BPP_PALETTE");
       if(VMatchEVar(env, "yes") == EMEV_Match) uecm = true;
  else if(VMatchEVar(env,  "no") == EMEV_Match) uecm = false;
  else if(env) VWarning(_NFL_, "Invalid value in environment variable \"VGE_VIDEO_CUSTOM_8BPP_PALETTE\"\nThe default system palette will be used");

  CreateMainWindow(un32WindowWidth, un32WindowHeight, utf8WindowTitle, NULL, uecm);
  CreateRenderImage(un32WindowWidth, un32WindowHeight);
  CreateRenderBuffer((un32RBuffWidth == 0) ? un32WindowWidth : un32RBuffWidth,
                     (un32RBuffHeight == 0) ? un32WindowHeight : un32RBuffHeight);
}

void CVideoEnv::CreateMainWindow(uint32_t un32Width, uint32_t un32Height,
                                 const char *utf8WindowTitle,
                                 const char *pcWindowXPMIconFile,
                                 bool bUseExclusiveColormap,
                                 const char *pcDisplayName)
{
  // Connect to the X server
  char *pcDEnv = NULL;
  bool  bNULL  = false;
  if(pcDisplayName) {
    _tX11.ptDisplay = XOpenDisplay(pcDisplayName);
    if(_tX11.ptDisplay == NULL) {
      pcDEnv = getenv("DISPLAY");
      if(pcDEnv)
        VInfo(_NFL_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(\"%s\")", pcDisplayName, pcDEnv);
      else {
        VInfo(_NFL_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(NULL)", pcDisplayName);
        bNULL = true;
      }
    }
  }
  else {
    pcDEnv = getenv("DISPLAY");
    if(pcDEnv == NULL) bNULL = true;
  }
  if(pcDEnv) {
    _tX11.ptDisplay = XOpenDisplay(pcDEnv);
    if(_tX11.ptDisplay == NULL) {
      bNULL = true;
      VInfo(_NFL_, "XOpenDisplay(\"%s\") failed, trying XOpenDisplay(NULL)", pcDEnv);
    }
  }
  if(bNULL) _tX11.ptDisplay = XOpenDisplay(NULL);
  if(_tX11.ptDisplay == NULL)
    VError(VERR_XOpenDisplay, _NFL_, "Couldn't connect to the X server");
  XSynchronize(_tX11.ptDisplay, False);

  // Initialize internal atoms
  __intexp_X11InitInternalAtoms(_tX11.ptDisplay);

  // Trap fatal X I/O error
  XSetIOErrorHandler(_XFatalIO_);

  // Check for XShm extension
  // If you got an error message similar to :
  //   VID_X11/Environment.cc:252: error: `XShmQueryExtensionas' undeclared (first use this function)
  // it is likely that your X11 headers do not declare the prototype, so try to define the
  // prototype manually (put it after the #includes section of this file:
  //   extern int XShmQueryExtensionas(Display *);
#ifdef __USE_XSHM__
  if(!XShmQueryExtension(_tX11.ptDisplay))
    VError(VERR_NotSupported, _NFL_, "It seems that the X server does not support XShm extension\nPerhaps you are using \"libXext\" with a bogus XShm support ?\nTry to rebuild VortexGE without XShm");
#endif

  // Get defaults
  _tX11.n32Screen = DefaultScreen(_tX11.ptDisplay);
  _tX11.un32Depth = DefaultDepth(_tX11.ptDisplay, _tX11.n32Screen);
  if((_tX11.un32Depth !=  8) &&
     (_tX11.un32Depth != 16) &&
     (_tX11.un32Depth != 24))
    VError(VERR_NotSupported, _NFL_, "Current display depth (%d bpp) is not supported", _tX11.un32Depth);
  if(_tX11.un32Depth ==  8)
#ifdef __USE_DITHERING_FOR_8BPP__
    VWarning(_NFL_, "Current display depth (8 bpp) is too low\nIt is recommended to use at least 16 bpp for better appearance");
#else
    VWarning(_NFL_, "Current display depth (8 bpp) is too low\nIt is recommended to use at least 16 bpp for much better appearance");
#endif
  _tX11.ptVisual    = DefaultVisual(_tX11.ptDisplay, _tX11.n32Screen);
  _tX11.tRootWindow = RootWindow(_tX11.ptDisplay, _tX11.n32Screen);
  _tX11.bRGB555     = !((_tX11.ptVisual->red_mask   == 0xf800) &&
                        (_tX11.ptVisual->green_mask == 0x07e0) &&
                        (_tX11.ptVisual->blue_mask  == 0x001f));
  _tX11.scrWidth    = DisplayWidth(_tX11.ptDisplay, _tX11.n32Screen);
  _tX11.scrHeight   = DisplayHeight(_tX11.ptDisplay, _tX11.n32Screen);

  // Get the size of the working area (desktop geometry)
  __intexp_X11GetDesktopGeometry(_tX11.ptDisplay, _tX11.n32Screen,
                                 _tX11.desktopX, _tX11.desktopY,
                                 _tX11.desktopW, _tX11.desktopH);

  // Create colormap
  _CreateColorMap_(bUseExclusiveColormap);

  // For 8 bpp only : generate palette table & color translation LUT
  if(_tX11.un32Depth == 8) {
    // Query colors from the colormap
    XColor tColors[256];
    for(ssize_t c = 0; c < 256; ++c) tColors[c].pixel = c;
    XQueryColors(_tX11.ptDisplay, _tX11.tColormap, tColors, 256);

    // Allocate memory for the palette table
    _tRI.pun16Pal = reinterpret_cast<uint16_t*>(malloc(256 * 4 * sizeof(uint16_t)));
    if(_tRI.pun16Pal == NULL)
      VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for palette table");

    // Generate the palette table
    uint16_t *pun16Tmp = _tRI.pun16Pal;
    for(size_t c = 0; c < 256; ++c) {
      *pun16Tmp++ = tColors[c].flags;
      *pun16Tmp++ = tColors[c].red;
      *pun16Tmp++ = tColors[c].green;
      *pun16Tmp++ = tColors[c].blue;
    }

    // Initialize the color translation LUT
    _InitColorLUT_();
  }

  // Create GC
  _tX11.tMainGC = XCreateGC(_tX11.ptDisplay, _tX11.tRootWindow, 0, NULL);
  if(_tX11.tMainGC < 0) {
    _tX11.tMainGC = 0;
    VError(VERR_XCreateGC, _NFL_, "XCreateGC failed");
  }

  // Initialize the main window attributes
  XSetWindowAttributes tWA;
  memset(&tWA, 0, sizeof(tWA));
  tWA.border_pixel     = _tX11.un32WhitePixel;
  tWA.background_pixel = _tX11.un32BlackPixel;
  tWA.colormap         = _tX11.tColormap;

  // Get the window frame size by creating a dummy window and then asking
  // the window manager (try it several times since with some window manager
  // and some version of X, it may fail several times)
  bool gwfzSuccess = false;
  VInfo(_NFL_, "Trying to ask the Window Manager for the standard window frame size ...");
  for(size_t wgfzc = 0; wgfzc < 5; ++wgfzc)
  {
    // Create a dummy window
    Window dw = XCreateWindow(_tX11.ptDisplay, _tX11.tRootWindow, 0, 0, 40, 30, 4,
                              _tX11.un32Depth, InputOutput, _tX11.ptVisual,
                              CWBorderPixel | CWBackPixel | CWColormap, &tWA);
    XSync(_tX11.ptDisplay, True);
    // Set the dummy window properties to macth the main window properties
    XSizeHints tz;
    tz.flags      = PMinSize | PMaxSize;
    tz.min_width  = tz.max_width  = 40;
    tz.min_height = tz.max_height = 30;
    XSetStandardProperties(_tX11.ptDisplay, dw, NULL, NULL, None, NULL, 0, &tz);
    Atom ta = XInternAtom(_tX11.ptDisplay, "_NET_WM_WINDOW_TYPE_NORMAL", False);
    XChangeProperty(_tX11.ptDisplay, dw, xprv::_x11Atom_NET_WM_TYPE, XA_ATOM,
                    32, PropModeReplace, reinterpret_cast<u_char*>(&ta), 1);
    XSync(_tX11.ptDisplay, True);
    // Map the window and try to ask for the frame size (try it several times
    // since it can fail too)
    for(size_t wafzc = 0; wafzc < 5; ++wafzc) {
      // Map the window
      XMapWindow(_tX11.ptDisplay, dw);
      XMoveWindow(_tX11.ptDisplay, dw, -4000, -3000);
      XSync(_tX11.ptDisplay, True);
      // Get the frame size
      gwfzSuccess = __intexp_X11GetWindowFrameSize(_tX11.ptDisplay, dw,
                                                   _tX11.wFrameL, _tX11.wFrameR,
                                                   _tX11.wFrameT, _tX11.wFrameB);
      //VPrint("Attempt[%d,%d] : L=%02d R=%02d T=%02d B=%02d\n", wgfzc, wafzc, _tX11.wFrameL, _tX11.wFrameR, _tX11.wFrameT, _tX11.wFrameB);
      // Break if success
      if(gwfzSuccess) break;
      // Delay for some time if not success
      usleep(25000);
    }
    // Unmap the window
    XUnmapWindow(_tX11.ptDisplay, dw);
    XSync(_tX11.ptDisplay, True);
    // Destroy the window
    XDestroyWindow(_tX11.ptDisplay, dw);
    XSync(_tX11.ptDisplay, True);
    // Break if success
    if(gwfzSuccess) break;
    // Delay for some time if not success
    usleep(100000);
  }
  if(gwfzSuccess)
    VPrint("Success: left=%d, right=%d, top=%d, bottom=%d\n", _tX11.wFrameL, _tX11.wFrameR, _tX11.wFrameT, _tX11.wFrameB);
  else
    VPrint("Failed: the Window Manager does not support getting the window frame size\n");

  // Adjust the window size using the size of the working area and
  // the window frame size
  if(un32Width  >= _tX11.desktopW) _tX11.un32Width  = un32Width  = _tX11.desktopW - _tX11.wFrameL - _tX11.wFrameR;
  if(un32Height >= _tX11.desktopH) _tX11.un32Height = un32Height = _tX11.desktopH - _tX11.wFrameT - _tX11.wFrameB;

  // Set the window width and/or height to the the screen
  // width and/or height if zero(es)
  if(!un32Width)  un32Width  = _tX11.scrWidth;
  if(!un32Height) un32Height = _tX11.scrHeight;

  // Create the main window
  _tX11.un32Width  = un32Width;
  _tX11.un32Height = un32Height;
  _tX11.tMainWindow = XCreateWindow(_tX11.ptDisplay, _tX11.tRootWindow,
                                      0, 0, un32Width, un32Height, 4, _tX11.un32Depth,
                                      InputOutput, _tX11.ptVisual,
                                      CWBorderPixel | CWBackPixel | CWColormap,
                                      &tWA);

  // Create XOM
#ifdef __USE_XOM__
  setlocale(LC_ALL, "en_US.UTF-8");
  _tX11.tMainXOM = XOpenOM(_tX11.ptDisplay, NULL, NULL, NULL);
  if(!_tX11.tMainXOM) VError(VERR_XOpenOM, _NFL_, "XOpenOM() failed");
#endif

  // Create XIM
#ifdef __USE_XIM__
  setlocale(LC_ALL, "en_US.UTF-8");
  _tX11.tMainXIM = XOpenIM(_tX11.ptDisplay, NULL, NULL, NULL);
  if(!_tX11.tMainXIM) VError(VERR_XOpenIM, _NFL_, "XOpenIM() failed");
  _tX11.tMainXIC = XCreateIC(_tX11.tMainXIM,
                             XNInputStyle,   XIMPreeditNothing | XIMStatusNothing,
                             XNClientWindow, _tX11.tMainWindow,
                             NULL);
  if(!_tX11.tMainXIC) {
    _tX11.tMainXIC = XCreateIC(_tX11.tMainXIM,
                               XNInputStyle,   XIMPreeditNone | XIMStatusNone,
                               XNClientWindow, _tX11.tMainWindow,
                               NULL);
  }
  if(!_tX11.tMainXIC) VError(VERR_XOpenIC, _NFL_, "XOpenIC() failed");
  XSetICFocus(_tX11.tMainXIC);
#endif

  // Select the event mask
  const u_long eventMask = KeyPressMask    | KeyReleaseMask    |
                           ButtonPressMask | ButtonReleaseMask | PointerMotionMask    |
                           FocusChangeMask | ExposureMask      | VisibilityChangeMask |
                           StructureNotifyMask ;
#ifdef __USE_XIM__
  u_long xicEventMask;
  XGetICValues(_tX11.tMainXIC, XNFilterEvents, &xicEventMask, NULL);
  XSelectInput(_tX11.ptDisplay, _tX11.tMainWindow, eventMask | xicEventMask);
#else
  XSelectInput(_tX11.ptDisplay, _tX11.tMainWindow, eventMask);
#endif

  // Set the main window properties
  XSizeHints tSize;
  tSize.flags = PMinSize | PMaxSize;
  tSize.min_width  = tSize.max_width  = un32Width;
  tSize.min_height = tSize.max_height = un32Height;
  XSetStandardProperties(_tX11.ptDisplay, _tX11.tMainWindow,
                         NULL, NULL, None, NULL, 0, &tSize);
  SetWindowCaption(utf8WindowTitle);

  // Set the main window type
  Atom tType = XInternAtom(_tX11.ptDisplay, "_NET_WM_WINDOW_TYPE_NORMAL", False);
  XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_NET_WM_TYPE,
                  XA_ATOM, 32, PropModeReplace, reinterpret_cast<u_char*>(&tType), 1);

  // Tell the window manager that we want to be sent a message when the user
  // press the close (X) button. Without this, the window manager will just
  // directly destroy the window
  tType = xprv::_x11Atom_WM_DELETE_WINDOW;
  XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_PROTOCOLS,
                  XA_ATOM, 32, PropModePrepend, reinterpret_cast<u_char*>(&tType), 1);

  // Create and use the default cursors
  _CreateDefaultCursors_();
  XSync(_tX11.ptDisplay, True);
  UseDefaultArrowCursor();

  // Create and use the default fonts
  _CreateDefaultFonts_();
  UseDefaultSmallFont();

  // Create and set the main window icon
  bool bSet = false;
  if(pcWindowXPMIconFile) {
    try {
      _LoadXPMFromFile_(pcWindowXPMIconFile, &_tX11.tMainIcon, &_tX11.tMainIconMask);
      _SetWindowIcon_(&_tX11.tMainIcon, &_tX11.tMainIconMask);
      bSet = true;
    }
    catch(...) {}
  }
  if(!bSet) SetWindowIconToDefault();

  // Allocate shared memory for the main pixmap
#ifdef __USE_XSHM__
  if(_tX11.un32Depth == 8)
    _tX11.tShmInfo.shmid = shmget(IPC_PRIVATE, 1024 + un32Width * un32Height, IPC_CREAT | IPC_EXCL | 0666);
  else if(_tX11.un32Depth == 16)
    _tX11.tShmInfo.shmid = shmget(IPC_PRIVATE, 1024 + un32Width * un32Height * 2, IPC_CREAT | IPC_EXCL | 0666);
  else
    _tX11.tShmInfo.shmid = shmget(IPC_PRIVATE, 1024 + un32Width * un32Height * 4, IPC_CREAT | IPC_EXCL | 0666);
  if(_tX11.tShmInfo.shmid == -1)
    VError(VERR_ShmGet, _NFL_, "Couldn't create a new shared memory segment for main pixmap");
  _tX11.tShmInfo.shmaddr = reinterpret_cast<char*>(shmat(_tX11.tShmInfo.shmid, NULL, 0));
  if(_tX11.tShmInfo.shmaddr == reinterpret_cast<char*>(-1))
    VError(VERR_ShmAt, _NFL_, "Couldn't attach the shared memory segment for main pixmap");
  //_tX11.tShmInfo.shmaddr = reinterpret_cast<char*>(reinterpret_cast<uintptr_t>(_tX11.tShmInfo.shmaddr + 32) &
  //                                                 ~static_cast<uintptr_t>(31) );
  _tX11.tShmInfo.readOnly = false;
  _TRAP_XSHM_ATTACH_ERROR_BEGIN_();
    XShmAttach(_tX11.ptDisplay, &_tX11.tShmInfo);
    XSync(_tX11.ptDisplay, True);
  _TRAP_XSHM_ATTACH_ERROR_END_(_tX11);
#endif

  // Create main pixmap
#ifdef __USE_XSHM__
  _tX11.tMainPixmap = XShmCreatePixmap(_tX11.ptDisplay, _tX11.tMainWindow,
                                       _tX11.tShmInfo.shmaddr, &_tX11.tShmInfo,
                                       un32Width, un32Height, _tX11.un32Depth);
#else
  _tX11.tMainPixmap = XCreatePixmap(_tX11.ptDisplay, _tX11.tMainWindow,
                                    un32Width, un32Height, _tX11.un32Depth);
#endif
  XSync(_tX11.ptDisplay, True);

  // Save the default style
  static const uint32_t aun32StpData[32] = { 0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,
                                             0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,
                                             0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,
                                             0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555,0xAAAAAAAA,0x55555555 };
  _tX11Style.ptFont           = _tX11.ptDefFontSmall;
  _tX11Style.ptFontSet        = _tX11.ptDefFontSetSmall;
  _tX11Style.xOutline         = CreateColor(  0,   0,   0);
  _tX11Style.xLightBorder     = CreateColor(255, 255, 255);
  _tX11Style.xDarkBorder      = CreateColor( 96,  96,  96);
  _tX11Style.xFace            = CreateColor(164, 164, 164);
  _tX11Style.xFaceLocked      = CreateColor(196, 172, 172);
  _tX11Style.xFaceSelect      = CreateColor(128, 128, 255);
  _tX11Style.xText            = CreateColor(  0,   0,   0);
  _tX11Style.xHotPress        = CreateColor( 64,  64, 192);
  _tX11Style.xIconBorder      = CreateColor(128, 128, 128);
  _tX11Style.xIconBackground  = CreateColor(  0,   0,   0);
  _tX11Style.xIconInformation = CreateColor(  0,   0, 255);
  _tX11Style.xIconQuestion    = CreateColor(  0, 255,   0);
  _tX11Style.xIconWarning     = CreateColor(255, 255,   0);
  _tX11Style.xIconError       = CreateColor(255,   0,   0);
  _tX11Style.xFillStyle       = (_tX11.un32Depth <= 8) ? FILL_SOLID : FILL_GRADIENT;
  _tX11CS.ptFont              = _tX11Style.ptFont;
  _tX11CS.ptFontSet           = _tX11Style.ptFontSet;
  _tX11CS.xForeground         = _tX11Style.xLightBorder;
  _tX11CS.xFillStyle          = _tX11Style.xFillStyle;
  memcpy(_tX11Style.aun32SPat, aun32StpData, _STIPPLE_ARRAY_SIZE_);

  // Clear the main pixmap
  SetFillStyle(FILL_SOLID);
  XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, _tX11.un32BlackPixel);
  XSetBackground(_tX11.ptDisplay, _tX11.tMainGC, _tX11.un32BlackPixel);
  XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC, 0, 0, un32Width, un32Height);
  XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, _tX11.un32WhitePixel);
  SetFillStyle(_tX11Style.xFillStyle);

  // Create and use the stiple pixmap
  SetStipplePattern(_tX11Style.aun32SPat);

  // Display and center the main window in screen
  XMapRaised(_tX11.ptDisplay, _tX11.tMainWindow);
  MoveCenterWindow();

  // Set auto repeat
  AutoRepeatOn();
  XSync(_tX11.ptDisplay, True);

  // OK !
  _tX11.bInitialized = true;
  _bActive           = true;
  _bVisible          = true;
}

void CVideoEnv::CreateRenderImage(uint32_t un32Width, uint32_t un32Height)
{
  // Temporary variable
  uint32_t un32Size;

  // Check
  vassert(_tX11.bInitialized,
          "The main window must be initialized properly before creating its render image");

  // Calculate image size
  _tRI.un32Width  = ((un32Width > 0) && (un32Width <  _tX11.un32Width)) ? un32Width  : _tX11.un32Width;
  _tRI.un32Height = ((un32Height > 0) && (un32Height <  _tX11.un32Height)) ? un32Height : _tX11.un32Height;
  _tRI.un32Dim    = _tRI.un32Width * _tRI.un32Height;
  if(_tX11.un32Depth == 8)
    un32Size = _tRI.un32Width * _tRI.un32Height;
  else if(_tX11.un32Depth == 16)
    un32Size = _tRI.un32Width * _tRI.un32Height * 2;
  else
    un32Size = _tRI.un32Width * _tRI.un32Height * 4;
  _tRI.un32Size = un32Size;

  // Allocate memory for render image using XShm
#ifdef __USE_XSHM__
  _tRI.tShmInfo.shmid = shmget(IPC_PRIVATE, un32Size, IPC_CREAT | IPC_EXCL | 0666);
  if(_tRI.tShmInfo.shmid == -1)
    VError(VERR_ShmGet, _NFL_, "Couldn't create a new shared memory segment for render image");
  _tRI.tShmInfo.shmaddr = reinterpret_cast<char*>(shmat(_tRI.tShmInfo.shmid, NULL, 0));
#ifdef _ALIGN_BUFF_MEM_
  if(uintptr_t(_tRI.tShmInfo.shmaddr) & uintptr_t(0xF))
    VError(VERR_NotSupported, _NFL_, "The attached shared memory address is not 16-byte aligned");
#endif
  if(_tRI.tShmInfo.shmaddr == reinterpret_cast<char*>(-1))
    VError(VERR_ShmAt, _NFL_, "Couldn't attach the shared memory segment for render image");
  _tRI.tShmInfo.readOnly = false;
  _TRAP_XSHM_ATTACH_ERROR_BEGIN_();
    XShmAttach(_tX11.ptDisplay, &_tRI.tShmInfo);
    XSync(_tX11.ptDisplay, True);
  _TRAP_XSHM_ATTACH_ERROR_END_(_tRI);
  _tRI.pucImage = reinterpret_cast<u_char*>(_tRI.tShmInfo.shmaddr);
  // Allocate memory for render image without using XShm
#else
  _tRI.pucImage = reinterpret_cast<u_char*>(abmmalloc(un32Size));
  if(_tRI.pucImage == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for image data");
#endif
  memset(_tRI.pucImage, 0, un32Size);

  // Create image
#ifdef __USE_XSHM__
  _tRI.ptImage = XShmCreateImage(_tX11.ptDisplay, _tX11.ptVisual,
                                   _tX11.un32Depth, ZPixmap,
                                   _tRI.tShmInfo.shmaddr, &_tRI.tShmInfo,
                                   _tRI.un32Width, _tRI.un32Height);
#else
  if(_tX11.un32Depth == 24)
    _tRI.ptImage = XCreateImage(_tX11.ptDisplay, _tX11.ptVisual, 24, ZPixmap, 0,
      reinterpret_cast<char*>(_tRI.pucImage), _tRI.un32Width,
      _tRI.un32Height, 8, _tRI.un32Width * (24 / 8) +
      _tRI.un32Width);
  else if(_tX11.un32Depth == 16)
    _tRI.ptImage = XCreateImage(_tX11.ptDisplay, _tX11.ptVisual, 16, ZPixmap, 0,
        reinterpret_cast<char*>(_tRI.pucImage), _tRI.un32Width,
        _tRI.un32Height, 8, _tRI.un32Width * (16 / 8));
  else
    _tRI.ptImage = XCreateImage(_tX11.ptDisplay, _tX11.ptVisual, 8, ZPixmap, 0,
      reinterpret_cast<char*>(_tRI.pucImage), _tRI.un32Width,
      _tRI.un32Height, 8, _tRI.un32Width * (8 / 8));
#endif
  if(_tRI.ptImage == NULL)
    VError(VERR_XCreateImage, _NFL_, "Couldn't create a new XImage");

  // OK !
  _tRI.bInitialized = true;
}

void CVideoEnv::CreateRenderBuffer(uint32_t un32Width, uint32_t un32Height)
{
  // Check
  vassert(_tRI.bInitialized,
          "Render image must be created before creating render buffer");

  // Calculate buffer size
  _tRB.un32Width  = ((un32Width > 0) && (un32Width <  _tRI.un32Width))  ? un32Width  : _tRI.un32Width;
  _tRB.un32Height = ((un32Height > 0) && (un32Height <  _tRI.un32Height)) ? un32Height : _tRI.un32Height;

  // Check buffer size
  _tRB.un32Size     = _tRB.un32Width * _tRB.un32Height * 4;
  _tRB.un32Size48   = _tRB.un32Size / 4 / 8;
  _tRB.un32Size4875 = (_tRB.un32Size48 * 3 / 4) & 0xFFFFFFFC;
  _tRB.un32Size4850 = (_tRB.un32Size48 * 2 / 4) & 0xFFFFFFFC;
  _tRB.un32Size4825 = (_tRB.un32Size48 * 1 / 4) & 0xFFFFFFFC;
  if((_tRB.un32Size48 * 4 * 8) != _tRB.un32Size)
    VError(VERR_InvalidParameter, _NFL_, "Render buffer size (un32Width X un32Height) must be divisible by 8");

  // Allocate memory for render buffer
  _tRB.pun8Buffer = reinterpret_cast<uint8_t*>(abmmalloc(_tRB.un32Size));
  if(_tRB.pun8Buffer == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for render buffer");
  memset(_tRB.pun8Buffer, 0, _tRB.un32Size);

  //  Allocate memory for Z buffer
  _tRB.pn32ZBuffer = reinterpret_cast<int32_t*>(abmmalloc(_tRB.un32Size));
  if(_tRB.pn32ZBuffer == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't allocate memory for Z buffer");
  memset(_tRB.pn32ZBuffer, 0, _tRB.un32Size);

  // Initialize some variables
  _tRB.un32SizeSL   = _tRB.un32Width * 4;
  _tRB.fWidth2      = float(_tRB.un32Width) / 2.0f;
  _tRB.fHeight2     = float(_tRB.un32Height) / 2.0f;
  _tRB.n32ZTrans    = 0;

  // Determine the update type
  if((_tRB.un32Width  == _tRI.un32Width ) &&
     (_tRB.un32Height == _tRI.un32Height)) {
    if(_tX11.un32Depth == 24)
      _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI24Bpp_;
    else if(_tX11.un32Depth == 16) {
      if(_tX11.bRGB555)
        _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI16Bpp555_;
      else
        _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI16Bpp565_;
    }
    else
      _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI8Bpp_;
  }
  else {
    if(_tX11.un32Depth == 24)
      _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI24BppScl_;
    else if(_tX11.un32Depth == 16) {
      if(_tX11.bRGB555)
        _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI16BppScl555_;
      else
        _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI16BppScl565_;
    }
    else
      _ConvertRB2RI_ = &CVideoEnv::_ConvertRB2RI8BppScl_;
  }

  // OK !
  _bAllInitialized = true;
}

void CVideoEnv::SetStyle(TStyle *ptNewStyle)
{
  // Check
  if(_tX11.un32Depth <= 8) {
    if(ptNewStyle->xFillStyle == FILL_GRADIENT) {
      ptNewStyle->xFillStyle = FILL_SOLID;
      VWarning(_NFL_, "Gradient fill style (FILL_GRADIENT) can't be used in 8 bpp display mode\nSolid fill style (FILL_SOLID) will be used");
    }
  }

  // Set style
  if(!ptNewStyle->ptFont)    ptNewStyle->ptFont    = _tX11.ptDefFontSmall;
  if(!ptNewStyle->ptFontSet) ptNewStyle->ptFontSet = _tX11.ptDefFontSetSmall;
  memcpy(&_tX11Style, ptNewStyle, sizeof(_tX11Style));

  // Activate style
  SetFont(_tX11Style.ptFont);
  SetUtf8Font(_tX11Style.ptFontSet);
  SetForegroundColor(_tX11Style.xLightBorder);
  SetFillStyle(_tX11Style.xFillStyle);
  SetStipplePattern(_tX11Style.aun32SPat);
}

void CVideoEnv::GetStyle(TStyle *ptStyle) const
{
  memcpy(ptStyle, &_tX11Style, sizeof(_tX11Style));
  memcpy(ptStyle->aun32SPat, _tX11CS.aun32SPat, _STIPPLE_ARRAY_SIZE_);
  ptStyle->ptFont     = _tX11CS.ptFont;
  ptStyle->ptFontSet  = _tX11CS.ptFontSet;
  ptStyle->xFillStyle = _tX11CS.xFillStyle;
}

void CVideoEnv::SetWindowIconToDefault()
{
  DelWindowIcon();
  _LoadXPMFromMemory_(_vgeIcon_xpm_, &_tX11.tMainIcon, &_tX11.tMainIconMask);
  return(_SetWindowIcon_(&_tX11.tMainIcon, &_tX11.tMainIconMask));
}

void CVideoEnv::SetWindowIcon(const char *xpmFileName)
{
  try {
    Pixmap tI = 0;
    Pixmap tM = 0;
    _LoadXPMFromFile_(xpmFileName, &tI, &tM);
    DelWindowIcon();
    _SetWindowIcon_(&tI, &tM);
    _tX11.tMainIcon     = tI;
    _tX11.tMainIconMask = tM;
  }
  catch(...) {}
}

void CVideoEnv::SetWindowIcon(const char *xpmDataBuffer[])
{
  try {
    Pixmap tI = 0;
    Pixmap tM = 0;
    _LoadXPMFromMemory_(xpmDataBuffer, &tI, &tM);
    DelWindowIcon();
    _SetWindowIcon_(&tI, &tM);
    _tX11.tMainIcon     = tI;
    _tX11.tMainIconMask = tM;
  }
  catch(...) {}
}

void CVideoEnv::DelWindowIcon()
{
  // Temporary variables
  XWMHints *ptWMH;

  // Remove the window icon
  ptWMH = XAllocWMHints();
  if(ptWMH) {
    // Set the icon to nothing
    ptWMH->flags       = IconPixmapHint | IconMaskHint;
    ptWMH->icon_pixmap = Pixmap(0);
    ptWMH->icon_mask   = Pixmap(0);
    XSetWMHints(_tX11.ptDisplay, _tX11.tMainWindow, ptWMH);
    XFree(ptWMH);
    // Free the main icon
    if(_tX11.tMainIcon)     XFreePixmap(_tX11.ptDisplay, _tX11.tMainIcon);
    if(_tX11.tMainIconMask) XFreePixmap(_tX11.ptDisplay, _tX11.tMainIconMask);
    _tX11.tMainIcon = _tX11.tMainIconMask = Pixmap(0);
  }
}

void CVideoEnv::SetWindowCaption(const char *utf8NormalCaption, const char *utf8IconicCaption)
{
  const char *nc = utf8NormalCaption ? utf8NormalCaption : "";
  const char *ic = utf8IconicCaption ? utf8IconicCaption : nc;

  const size_t nl = strlen(nc);
  const size_t il = strlen(ic);

  // Set / delete the normal caption
  if(nl) {
    XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_NAME, XA_STRING,
                    8, PropModeReplace, reinterpret_cast<const u_char*>(nc), nl);
    XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_NET_WM_NAME, XA_STRING,
                    8, PropModeReplace, reinterpret_cast<const u_char*>(nc), nl);
  }
  else {
    XDeleteProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_NAME);
    XDeleteProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_NET_WM_NAME);
  }

  // Set / delete the iconic caption
  if(il) {
    XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_ICON_NAME, XA_STRING,
                    8, PropModeReplace, reinterpret_cast<const u_char*>(ic), il);
    XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_NET_WM_ICON_NAME, XA_STRING,
                    8, PropModeReplace, reinterpret_cast<const u_char*>(ic), il);
  }
  else {
    XDeleteProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_ICON_NAME);
    XDeleteProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_NET_WM_ICON_NAME);
  }
}

void CVideoEnv::SetTransientFor(CVideoEnv *vEnv)
{
  if(vEnv) {
    Window win = vEnv->_tX11.tMainWindow;
    XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_TRANSIENT_FOR,
                    XA_WINDOW, 32, PropModeReplace, reinterpret_cast<const u_char*>(&win), 1);
  }
  else {
    XDeleteProperty(_tX11.ptDisplay, _tX11.tMainWindow, xprv::_x11Atom_WM_TRANSIENT_FOR);
  }
}

void CVideoEnv::AutoRepeatOn()
{
  XAutoRepeatOn(_tX11.ptDisplay);
  _tX11.bAutoRepeat = true;
}

void CVideoEnv::AutoRepeatOff()
{
  XAutoRepeatOff(_tX11.ptDisplay);
  _tX11.bAutoRepeat = false;
}

bool CVideoEnv::IsFullscreen() const
{
  return(_pcsFSOwner_ == this);
}

void CVideoEnv::SetFullscreen(bool bState)
{
  // Check
  if((_pcsFSHoldr_ != NULL) && (_pcsFSHoldr_ != this))
    VError(VERR_InUse, _NFL_, "Window %p still in the fullscreen state; only one window at a time can goes fullscreen", _pcsFSHoldr_);

  // No need to switch again if we own the fullscreen window and
  // the new requested state is also to switch to fullscreen
  if((_pcsFSOwner_ == this) && (bState)) return;

  // Switch to fullscreen
  if(bState) {
    // Set owner to this
    _pcsFSOwner_ = this;
    // Save the original video mode
    __intexp_XF86SaveMode_(_tX11.ptDisplay, _tX11.n32Screen, &_tSavedVideoMode_);
    // Switch video mode
    __intexp_XF86ModeInfo_ **pptModeList;
    int32_t                n32Tot, n32Idx;
    __intexp_XF86QueryVideoModes_(_tX11.ptDisplay, _tX11.n32Screen, &pptModeList, &n32Tot);
    n32Idx = __intexp_XF86FindNearestVideoMode_(_tX11.un32Width, _tX11.un32Height, pptModeList, n32Tot);
    __intexp_XF86SwitchToVideoMode_(_tX11.ptDisplay, _tX11.n32Screen, pptModeList[n32Idx]);
    __intexp_XF86FreeVideoModes_(pptModeList, n32Tot);
    // Change the window state
#define SET_WM_HINTS(HINTS)                                 \
  XChangeProperty(_tX11.ptDisplay, _tX11.tMainWindow,       \
                  wmha, wmha, sizeof(uint32_t) * 8,         \
                  PropModeReplace, (const uint8_t*) &HINTS, \
                  sizeof(HINTS) / sizeof(uint32_t))
    Atom wmha;
    if((wmha = XInternAtom(_tX11.ptDisplay, "_MOTIF_WM_HINTS", True)) != None) {
      const uint32_t MWM_HINT_DECORATIONS = uint32_t(1) << 1;
      struct SMotifWMHints {
        uint32_t m_Flags;
        uint32_t m_Functions;
        uint32_t m_Decorations;
        int32_t  m_InputMode;
        uint32_t m_Status;
      } MOTIF_hints = { MWM_HINT_DECORATIONS, 0, 0, 0, 0 };
      SET_WM_HINTS(MOTIF_hints);
    }
    else if((wmha = XInternAtom(_tX11.ptDisplay, "_WIN_HINTS", True)) != None) {
      int32_t GNOME_hints = 0;
      SET_WM_HINTS(GNOME_hints);
    }
    else if((wmha = XInternAtom(_tX11.ptDisplay, "KWM_WIN_DECORATION", True)) != None) {
      int32_t KDE_hints = 0;
      SET_WM_HINTS(KDE_hints);
    }
    else {
      // Nothing available, oh well
    }
    // Move the window to (0, 0) and resize it to fit as much as possible with the resolution
    XMoveResizeWindow(_tX11.ptDisplay, _tX11.tMainWindow, 0, 0, _tX11.un32Width, _tX11.un32Height);
    XRaiseWindow(_tX11.ptDisplay, _tX11.tMainWindow);
    // Grab the pointer to prevent WM from taking the focus or scrolling the virtual desktop
    XGrabPointer(_tX11.ptDisplay, _tX11.tMainWindow, False,
                 ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
                 GrabModeAsync, GrabModeAsync, _tX11.tMainWindow, None, CurrentTime);
    // Warp the pointer to make sure that the window fills the screen correctly
    XWarpPointer(_tX11.ptDisplay, None, _tX11.tMainWindow, 0, 0, 0, 0, 0, 0);
    XSync(_tX11.ptDisplay, false);
    XWarpPointer(_tX11.ptDisplay, None, _tX11.tMainWindow, 0, 0, 0, 0, _tX11.un32Width / 2, _tX11.un32Height / 2);
    XSync(_tX11.ptDisplay, false);
    _tEB.un32Xm = _tX11.un32Width / 2;
    _tEB.un32Ym = _tX11.un32Height / 2;
  }
  // Switch to windowed
  else {
    // Ungrab the pointer
    XUngrabPointer (_tX11.ptDisplay, CurrentTime);
    // Restore the video mode
    __intexp_XF86RestoreMode_(_tX11.ptDisplay, _tX11.n32Screen, &_tSavedVideoMode_);
    // Change the window state
    Atom wmha = XInternAtom(_tX11.ptDisplay, "_MOTIF_WM_HINTS", True);
    if(wmha == None) wmha = XInternAtom(_tX11.ptDisplay, "_WIN_HINTS", True);
    if(wmha == None) wmha = XInternAtom(_tX11.ptDisplay, "KWM_WIN_DECORATION", True);
    XDeleteProperty (_tX11.ptDisplay, _tX11.tMainWindow, wmha);
    // Move the window back to the center of the screen
    MoveCenterWindow();
    // Set owner to nothing
    _pcsFSOwner_ = NULL;
  }
}

void CVideoEnv::SetAlwaysOnTop(bool bState)
{
  _Set_NET_WM_STATE_(bState ? 1 : 0, "_NET_WM_STATE_ABOVE");
}

void CVideoEnv::MoveWindow(uint32_t un32XPos, uint32_t un32YPos)
{
  XMoveWindow(_tX11.ptDisplay, _tX11.tMainWindow, un32XPos, un32YPos);
  XSync(_tX11.ptDisplay, False);
}

void CVideoEnv::MoveCenterWindow(bool onDesktop)
{
  uint32_t px, py;

  if(onDesktop) {
    px = (_tX11.desktopW - _tX11.un32Width ) / 2 + _tX11.desktopX;
    py = (_tX11.desktopH - _tX11.un32Height) / 2 + _tX11.desktopY;
  }
  else {
    px = (_tX11.scrWidth - _tX11.un32Width) / 2;
    py = (_tX11.scrHeight - _tX11.un32Height) / 2;
  }

  if(_tX11.un32Width  >= _tX11.desktopW-_tX11.wFrameL-_tX11.wFrameR) px = _tX11.desktopX;
  if(_tX11.un32Height >= _tX11.desktopH-_tX11.wFrameT-_tX11.wFrameB) py = _tX11.desktopY;

  XMoveWindow(_tX11.ptDisplay, _tX11.tMainWindow, px, py);
  XSync(_tX11.ptDisplay, False);
}

void CVideoEnv::ClearEvent()
{
  memset(&_tEB, 0, sizeof(_tEB));
  _tEB.ptEvent = &_tEB.tEmptyEvent;
  XSync(_tX11.ptDisplay, True);
}

void CVideoEnv::DispatchEvent()
{
  // If visible use the nonblocking mode
  _if_likely(_bVisible) {
    // Process all events till empty
    while(_XPending_()) {
      XNextEvent(_tX11.ptDisplay, &_tEB.atEvents[_tEB.un8DPos]);
      if(!_HandleInternalEvents_(_tEB.atEvents[_tEB.un8DPos])) {
        if(_UsableEvent_(_tEB.atEvents[_tEB.un8DPos])) {
          _tEB.un8DPos++;
          if(_tEB.un8DPos >= _MAX_EVENTS_) _tEB.un8DPos = 0;
        }
      }
    }
    // This ugly hack is no longer needed since we now use pointer grab.
    //_MaintainViewport00_();
  }
  // If not visible use the blocking mode
  else
    BlockedDispatchEvent();
}

void CVideoEnv::BlockedDispatchEvent()
{
  XNextEvent(_tX11.ptDisplay, &_tEB.atEvents[_tEB.un8DPos]);
  if(!_HandleInternalEvents_(_tEB.atEvents[_tEB.un8DPos])) {
    if(_UsableEvent_(_tEB.atEvents[_tEB.un8DPos])) {
      _tEB.un8DPos++;
      if(_tEB.un8DPos >= _MAX_EVENTS_) _tEB.un8DPos = 0;
    }
  }
}

bool CVideoEnv::CheckEvent()
{
  if(_tEB.un8CPos == _tEB.un8DPos) {
    _tEB.ptEvent = &_tEB.tEmptyEvent;
    return(false);
  }
  _tEB.ptEvent = &_tEB.atEvents[_tEB.un8CPos];
  _tEB.un8CPos++;
  if(_tEB.un8CPos >= _MAX_EVENTS_) _tEB.un8CPos = 0;
  return(true);
}

void CVideoEnv::PutBackEvent()
{
  if(_tEB.un8CPos == 0)
    _tEB.un8CPos = _MAX_EVENTS_ - 1;
  else
  _tEB.un8CPos--;
}

void CVideoEnv::PopEvent(void **data)
{
  *data = NULL;

  void *p = malloc(sizeof(*_tEB.ptEvent));
  _if_unlikely(!p)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to pop (and save) the current event data");

  if(_tEB.un8CPos == _tEB.un8DPos)
    memcpy(p, &_tEB.tEmptyEvent, sizeof(*_tEB.ptEvent));
  else
    memcpy(p, &_tEB.atEvents[_tEB.un8CPos], sizeof(*_tEB.ptEvent));

  *data = p;

  _tEB.un8CPos++;
  if(_tEB.un8CPos >= _MAX_EVENTS_) _tEB.un8CPos = 0;
}

void CVideoEnv::PushEvent(void **data)
{
  _if_unlikely(!data) return;

  if(_tEB.un8CPos == 0)
    _tEB.un8CPos = _MAX_EVENTS_ - 1;
  else
  _tEB.un8CPos--;

   memcpy(&_tEB.atEvents[_tEB.un8CPos], *data, sizeof(*_tEB.ptEvent));

  free(*data);
  *data = NULL;
}

void CVideoEnv::FreeStoredEvent(void **data) const
{
  free(*data);
  *data = NULL;
}

bool CVideoEnv::_XPending_() const
{
  // Check for queued events
  XFlush(_tX11.ptDisplay);
  if(XEventsQueued(_tX11.ptDisplay, QueuedAlready)) return(true);

  // If not, check whether the file descriptor has changed status
  fd_set  fds;
  int32_t x11FD = ConnectionNumber(_tX11.ptDisplay);
  FD_ZERO(&fds);
  FD_SET(x11FD, &fds);

  struct timeval tTimeOut = {0, 0};
  if(select(x11FD + 1, &fds, NULL, NULL, &tTimeOut) == 1)
    if(XPending(_tX11.ptDisplay)) return(true);

  // No queued events found
  return(false);
}

bool CVideoEnv::_HandleInternalEvents_(XEvent &tEvent)
{
  // Flags
  bool bInt = false;

  // Handle events
  if(tEvent.type == MotionNotify) {
    _tEB.un32Xm = tEvent.xmotion.x;
    _tEB.un32Ym = tEvent.xmotion.y;
  }
  else if(tEvent.type == FocusIn) {
    _hie_bFocus = true;
    tEvent.type = _vge_FocusChange;
#ifdef __USE_XIM__
    XSetICFocus(_tX11.tMainXIC);
#endif
  }
  else if(tEvent.type == FocusOut) {
    _hie_bFocus = false;
    tEvent.type = _vge_FocusChange;
#ifdef __USE_XIM__
    XUnsetICFocus(_tX11.tMainXIC);
#endif
  }
  else if(tEvent.type == MapNotify) {
    _hie_bMap = true;
    bInt      = true;
  }
  else if(tEvent.type == UnmapNotify) {
    _hie_bMap = false;
    bInt      = true;
  }
  else if(tEvent.type == VisibilityNotify) {
    _hie_bVisible = (tEvent.xvisibility.state != VisibilityFullyObscured);
    bInt          = true;
  }
  else if(tEvent.type == DestroyNotify) {
    _hie_bMap = false;
    bInt      = true;
  }
  else if(tEvent.type == SelectionClear) {
    // Lock first
    CMutex::StdLocker lock_this(xprv::_selectionMutex);
    // Process the request
    bInt = true;
    xprv::_x11SelectionOwner     = 0;
    xprv::_x11SelectionRequestor = 0;
    xprv::_x11SelectionDataToGet = CClipboard::CDT_Null;
    xprv::_x11SelectionDataType  = CClipboard::CDT_Null;
    xprv::_x11SelectionData_String.clear();
  }
  else if(tEvent.type == SelectionNotify) {
    bInt = true;
    if(tEvent.xselection.property && xprv::_x11SelectionRequestor==_tX11.tMainWindow) {
      // Lock first
      CMutex::StdLocker lock_this(xprv::_selectionMutex);
      // Variables used to get the data
      long    bytesread = 0;
      Atom    actual    = 0;
      int     format    = 0;
      u_long  count     = 0;
      u_long  remaining = 0;
      u_char *portion   = 0;
      // Flag to indicate if the data can be acquired
      bool acquired = false;
      // String data ?
      if(xprv::_x11SelectionDataToGet == CClipboard::CDT_String) {
        // Clear the old string data
        xprv::_x11SelectionData_String.clear();
        // Get the new string data
        const size_t lm = 65536;
        for(;;) {
          if(XGetWindowProperty(_tX11.ptDisplay,
                                tEvent.xselection.requestor,
                                tEvent.xselection.property,
                                bytesread / 4, lm, 1, 0,
                                &actual, &format, &count, &remaining, &portion)) break;
          try        { xprv::_x11SelectionData_String += reinterpret_cast<char*>(portion); }
          catch(...) { XFree(portion); break; }
          XFree(portion);
          bytesread += (count * format / 8);
          if(!remaining) break;
        }
        acquired = true;
      }
      // Invalid data, just erase the data from the window's property
      else {
        XDeleteProperty(_tX11.ptDisplay,
                        tEvent.xselection.requestor,
                        tEvent.xselection.property);
      }
      // Process further if the data was acquired
      if(acquired) {
        // Set flags
        xprv::_x11SelectionDataType  = xprv::_x11SelectionDataToGet;
        xprv::_x11SelectionDataToGet = CClipboard::CDT_Null;
        // Send an event to indicate that the selection has been acquired
        XEvent tXEvent;
        memset(&tXEvent, 0, sizeof(tXEvent));
        tXEvent.xclient.type         = ClientMessage;
        tXEvent.xclient.message_type = xprv::_x11Atom_WM_PROTOCOLS;
        tXEvent.xclient.display      = _tX11.ptDisplay;
        tXEvent.xclient.window       = _tX11.tMainWindow;
        tXEvent.xclient.format       = 32;
        tXEvent.xclient.data.l[0]    = _vge_PasteData;
        tXEvent.xclient.data.l[1]    = xprv::_x11SelectionDataType;
        XSendEvent(_tX11.ptDisplay, _tX11.tMainWindow, False, 0, &tXEvent);
        XSync(_tX11.ptDisplay, False);
      }
    }
  }
  else if(tEvent.type == SelectionRequest) {
    // Lock first
    CMutex::StdLocker lock_this(xprv::_selectionMutex);
    // Process the request
    bInt = true;
    XSelectionEvent e;
    e.type      = SelectionNotify;
    e.requestor = tEvent.xselectionrequest.requestor;
    e.selection = tEvent.xselectionrequest.selection;
    e.target    = tEvent.xselectionrequest.target;
    e.time      = tEvent.xselectionrequest.time;
    e.property  = tEvent.xselectionrequest.property;
    if(e.target == xprv::_x11Atom_TARGETS) {
      // Determine the data type to be copied
      Atom type = 0;
      if(xprv::_x11SelectionDataType == CClipboard::CDT_String) type = XA_STRING;
      // Set the property
      if(type) XChangeProperty(_tX11.ptDisplay, e.requestor, e.property,
                               XA_ATOM, sizeof(Atom) * 8, PropModeReplace,
                               reinterpret_cast<u_char*>(&type), sizeof(Atom));
    }
    else if(xprv::_x11SelectionDataType == CClipboard::CDT_String) {
      const u_char *dt = reinterpret_cast<const u_char*>(xprv::_x11SelectionData_String.data());
            size_t  sz = xprv::_x11SelectionData_String.length();
      const size_t  lm = 65536;
      // Copy chunk by chunk if the data to be copied is larger that the size limit
      if(sz > lm) {
        // Copy the first chunk
        XChangeProperty(_tX11.ptDisplay, e.requestor, e.property, e.target, 8,
                        PropModeReplace, dt, lm);
        sz -= lm; dt += lm;
        // Copy the other chunks
        while(sz > lm) {
          XChangeProperty(_tX11.ptDisplay, e.requestor, e.property, e.target, 8,
                          PropModeAppend, dt, lm);
          sz -= lm; dt += lm;
        }
        // Copy the last chunk (if there is any)
        if(sz) XChangeProperty(_tX11.ptDisplay, e.requestor, e.property, e.target, 8,
                               PropModeAppend, dt, sz);
      }
      // Otherwise, copy it with a single pass
      else {
        XChangeProperty(_tX11.ptDisplay, e.requestor, e.property, e.target, 8,
                        PropModeReplace, dt, sz);
      }
    }
    else {
      e.property = 0;
    }
    XSendEvent(_tX11.ptDisplay, e.requestor, False, 0, reinterpret_cast<XEvent*>(&e));
  }
  else if(tEvent.type == ClientMessage) {
    if(tEvent.xclient.message_type == xprv::_x11Atom_WM_PROTOCOLS) {
      if(Atom(tEvent.xclient.data.l[0]) == xprv::_x11Atom_WM_DELETE_WINDOW) {
        tEvent.type = _vge_CloseWindow;
      }
      else if(tEvent.xclient.data.l[0] == _vge_PasteData) {
        tEvent.type = _vge_PasteData;
      }
    }
  }

  // Adjust the application states
  _bVisible = _hie_bMap & _hie_bVisible;
  _bActive  = _hie_bFocus & _hie_bVisible;
  if(_bActive) {
    if(_tX11.bAutoRepeat)
      XAutoRepeatOn(_tX11.ptDisplay);
    else
      XAutoRepeatOff(_tX11.ptDisplay);
  }
  else
    XAutoRepeatOn(_tX11.ptDisplay);
  _ToggleFullscreen_(_bActive);

  // Filter the event
#ifdef __USE_XIM__
  if(XFilterEvent(&tEvent, 0) == True) return(true);
#endif

  // Tell whether the internal event is completely handled
  // and there is no need to queue it
  return(bInt);
}

EKeySymbol CVideoEnv::GetKeySym() const
{
  // Temporary variables
  KeySym   tKS;
  uint32_t un32KS;

  // Get the X11 key symbol
  tKS = XKeycodeToKeysym(_tX11.ptDisplay, _tEB.ptEvent->xkey.keycode, 0);

  // Translate the symbol
  if(tKS) {
    switch(tKS & 0xFFFFFF00) {
      case 0x00000000: // Latin 1
      case 0x00000100: // Latin 2
      case 0x00000200: // Latin 3
      case 0x00000300: // Latin 4
      case 0x00000400: // Katakana
      case 0x00000500: // Arabic
      case 0x00000600: // Cyrillic
      case 0x00000700: // Greek
      case 0x00000C00: // Hebrew
      case 0x00000D00: // Thai
      case 0x00000E00: // Korean
      case 0x00001400: // Armenian
      case 0x00001500: // Georgian
      case 0x00001600: // Caucasius
      case 0x00001E00: // Vietnamese
        un32KS = tKS & 0x000000FF;
        if((un32KS >= 'A') && (un32KS <= 'Z')) un32KS += ('a' - 'A');
        return(EKeySymbol(un32KS));
      case 0x0000FE00:
        switch(tKS & 0x0000FFFF) {
          case XK_ISO_Level3_Shift: return(VKS_MODE);
          case XK_dead_circumflex : return(VKS_CARET);
          default                 : return(VKS_UNKNOWN);
        }
      case 0x0000FF00:
        return(CVideoEnv::_VKS[tKS & 0x000000FF]);
      default:
        return(VKS_UNKNOWN);
    }
  }
  return(VKS_UNKNOWN);
}

uint32_t CVideoEnv::GetKeyString(char *retBuff, uint32_t retBuffSize, uint16_t *retUCS) const
{
#ifdef __USE_XIM__
  KeySym   ksym;
  Status   status;
  uint32_t bfsz = Xutf8LookupString(_tX11.tMainXIC, &_tEB.ptEvent->xkey, retBuff, retBuffSize, &ksym, &status);

  if((status != XLookupBoth) && (status != XLookupChars)) {
    if(retUCS) *retUCS = 0;
    return(0);
  }

  if(retUCS) {
    CUString utf32(retBuff);
    if(utf32.length())
      *retUCS = utf32[0];
    else
      *retUCS = __intexp_XKeySym2UCS_(ksym);
  }
  return(bfsz);
#else
  KeySym   ksym;
  uint32_t bfsz = XLookupString(&_tEB.ptEvent->xkey, retBuff, retBuffSize, &ksym, NULL);

  if(retUCS) *retUCS = __intexp_XKeySym2UCS_(ksym);
  return(bfsz);
#endif
}

TFont *CVideoEnv::CreateFont(const CFontDesc &csFont) const
{
  // Preparations
  string         name1, name2, name3;
  vector<string> fName, pSize;
  _PrepareCreateFont_(csFont, name1, name2, name3, fName, pSize);

  // Create the font
  TFont *fnt = _nothrow_new TFont;
  if(!fnt)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new font");
  fnt->_ul = csFont.GetUnderline();

  for(vector<string>::const_iterator f = fName.begin(); f != fName.end(); f++) {
    for(vector<string>::const_iterator p = pSize.begin(); p != pSize.end(); p++) {
      // Combine the name
      string name = name1;
      name += *f;
      name += name2;
      name += *p;
      name += name3;
      // Create the font
      *fnt = XLoadQueryFont(_tX11.ptDisplay, name.c_str());
      if(fnt->_fs) break;
    }
    if(fnt->_fs) break;
  }

  if(!fnt->_fs) {
    string name = name1;
    name += csFont.GetFamily();
    name += name2;
    name += csFont.GetPointSize();
    name += name3;
    VError(VERR_XLoadQueryFont, _NFL_, "Couldn't load font \"%s\"", name.c_str());
  }

  // Return the font
  return(fnt);
}

void CVideoEnv::FreeFont(TFont *ptFont)
{
  if((ptFont == _tX11.ptDefFontSmall) || (ptFont == _tX11.ptDefFontBig)) return;
  XFreeFont(_tX11.ptDisplay, *ptFont);
  if(_tX11CS.ptFont == ptFont) SetFont(_tX11.ptDefFontSmall);
  delete ptFont;
}

void CVideoEnv::SetFont(TFont *ptFont)
{
  // If the new font is the same as the current font, just return (no need to set)
  if(_tX11CS.ptFont == ptFont) return;

  // Set the new font
  _tX11CS.ptFont = ptFont;
  XSetFont(_tX11.ptDisplay, _tX11.tMainGC, (*_tX11CS.ptFont)->fid);
}

void CVideoEnv::GetUtf8TextSize(uint32_t &w, uint32_t &h, const string &csText) const
{
  XRectangle tl;
  Xutf8TextExtents(*_tX11CS.ptFontSet, csText.data(), csText.length(), NULL, &tl);
  w = tl.width;
  h = tl.height;
}

void CVideoEnv::GetUtf8TextGeometry(int32_t &x, int32_t &y, uint32_t &w, uint32_t &h, const string &csText) const
{
  XRectangle tl;
  Xutf8TextExtents(*_tX11CS.ptFontSet, csText.data(), csText.length(), NULL, &tl);
  x = tl.x;
  y = tl.y;
  w = tl.width;
  h = tl.height;
}

TFontSet *CVideoEnv::CreateUtf8Font(const CFontDesc &csFont) const
{
  // Preparations
  string         name1, name2, name3;
  vector<string> fName, pSize;
  _PrepareCreateFont_(csFont, name1, name2, name3, fName, pSize);

  // Create the font
  TFontSet *fnt = _nothrow_new TFontSet;
  if(!fnt)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new font set");
  fnt->_ul = csFont.GetUnderline();

  for(vector<string>::const_iterator f = fName.begin(); f != fName.end(); f++) {
    for(vector<string>::const_iterator p = pSize.begin(); p != pSize.end(); p++) {
      // Combine the name
      string name = name1;
      name += *f;
      name += name2;
      name += *p;
      name += name3;
      // Create the UTF-8 font (using "en_US.UTF-8" as the C locale)
      setlocale(LC_ALL, "en_US.UTF-8");
#ifdef __USE_XOM__
      *fnt = XCreateOC(_tX11.tMainXOM, XNBaseFontName, name.c_str(), NULL);
#else
      int32_t n32Dummy;
      char   *pcDummy, **ppcDummy;
      *fnt = XCreateFontSet(_tX11.ptDisplay, name.c_str(), &ppcDummy, &n32Dummy, &pcDummy);
      XFreeStringList(ppcDummy);
#endif
      if(fnt->_fs) break;
    }
    if(fnt->_fs) break;
  }

  if(!fnt->_fs) {
    string name = name1;
    name += csFont.GetFamily();
    name += name2;
    name += csFont.GetPointSize();
    name += name3;
    VError(VERR_XCreateFontSet, _NFL_, "Couldn't create UTF-8 fontset \"%s\"", name.c_str());
  }

  // Return the font
  return(fnt);
}

void CVideoEnv::FreeUtf8Font(TFontSet *ptFontSet)
{
  if((ptFontSet == _tX11.ptDefFontSetSmall) || (ptFontSet == _tX11.ptDefFontSetBig)) return;
#ifdef __USE_XOM__
  XDestroyOC(*ptFontSet);
#else
  XFreeFontSet(_tX11.ptDisplay, *ptFontSet);
#endif
  if(_tX11CS.ptFontSet == ptFontSet) SetUtf8Font(_tX11.ptDefFontSetSmall);
  delete ptFontSet;
}

void CVideoEnv::SetUtf8Font(TFontSet *ptFontSet)
{
  _tX11CS.ptFontSet = ptFontSet;
}

XColor24 CVideoEnv::CreateColor(uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  if(_tX11.un32Depth == 24) return(X11_RGB24(un8R, un8G, un8B));
  if(_tX11.un32Depth == 16) return(X11_RGB16(un8R, un8G, un8B, _tX11.bRGB555));
#ifdef __USE_DITHERING_FOR_8BPP__
  return(_tRI.pucLUT[(un8R & 0xE0) | ((un8G & 0xE0) >> 3)| ((un8B & 0xC0) >> 6)]);
#else
  return(_tRI.pucLUT[(uint32_t(un8R & 0xF8) << 7) |
                       (uint32_t(un8G & 0xF8) << 2) |
                        uint32_t(un8B         >> 3)]);
#endif
}

void CVideoEnv::ExtractColor(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B, XColor24 xColor)
{
  // Extract from 8bpp
  if(_tX11.un32Depth == 8) {
    uint32_t i = (uint32_t(xColor) << 2) + 1;
    un8R = _tRI.pun16Pal[i++] >> 8;
    un8G = _tRI.pun16Pal[i++] >> 8;
    un8B = _tRI.pun16Pal[i++] >> 8;
  }
  // Extract from 16bpp
  else if(_tX11.un32Depth == 16) {
    if(_tX11.bRGB555) {
      un8R = ((xColor >> 10) & 0x0000001F) << 3;
      un8G = ((xColor >>  5) & 0x0000001F) << 3;
      un8B = ((xColor >>  0) & 0x0000001F) << 3;
    }
    else {
      un8R = ((xColor >> 11) & 0x0000001F) << 3;
      un8G = ((xColor >>  5) & 0x0000003F) << 2;
      un8B = ((xColor >>  0) & 0x0000001F) << 3;
    }
  }
  //  Extract from 24bpp
  else {
    un8R = (xColor >> 16) & 0x000000FF;
    un8G = (xColor >>  8) & 0x000000FF;
    un8B = (xColor >>  0) & 0x000000FF;
  }
}

void CVideoEnv::SetForegroundColor(XColor24 xColor)
{
  if(_tX11CS.xForeground == xColor) return;
  _tX11CS.xForeground = xColor;
  XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, _tX11CS.xForeground);
}

void CVideoEnv::SetFillStyle(XFillStyle xFillStyle)
{
  if(_tX11CS.xFillStyle == xFillStyle) return;
  _tX11CS.xFillStyle = xFillStyle;
  if(_tX11CS.xFillStyle == FILL_STIPPLED)
    XSetFillStyle(_tX11.ptDisplay, _tX11.tMainGC, FillStippled);
  else
    XSetFillStyle(_tX11.ptDisplay, _tX11.tMainGC, FillSolid);
}

void CVideoEnv::SetStipplePattern(const uint32_t aun32SPat[_STIPPLE_SIZE_])
{
  memcpy(_tX11CS.aun32SPat, aun32SPat, _STIPPLE_ARRAY_SIZE_);

  // Set the stipple pixmap
  Pixmap tStipple = XCreatePixmapFromBitmapData(_tX11.ptDisplay, _tX11.tMainWindow,
                                                reinterpret_cast<char*>(const_cast<uint32_t*>(aun32SPat)),
                                                32, _STIPPLE_SIZE_, 1L, 0L, 1);
  XSetStipple(_tX11.ptDisplay, _tX11.tMainGC, tStipple);
  XFreePixmap(_tX11.ptDisplay, tStipple);
  XSync(_tX11.ptDisplay, True);
}

void CVideoEnv::DrawText(int32_t n32X, int32_t n32Y, const string &csText, ETextAlignment eAlignment)
{
  // Convert to C-style string
  const char *text = csText.c_str();
  uint32_t     len = csText.length();

  // Center on screen horizontally
  int32_t pX;
  if(eAlignment == ETA_CenterScreen)
    pX = (_tX11.un32Width - XTextWidth(*_tX11CS.ptFont, text, len)) >> 1;
  // Center on the given X coordinate horizontally
  else if(eAlignment == ETA_Center)
    pX = n32X - (XTextWidth(*_tX11CS.ptFont, text, len) >> 1);
  // Right aligned on the given X coordinate horizontally
  else if(eAlignment == ETA_Right)
    pX = n32X - XTextWidth(*_tX11CS.ptFont, text, len);
  // Left aligned on the given X coordinate horizontally
  else
    pX = n32X;

  // Draw the text
  XDrawString(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC, pX, n32Y, text, len);
}

void CVideoEnv::DrawTextML(int32_t n32X, int32_t n32Y, const string &csText)
{
#if 0
  // For convenience
  const string           &str    = csText;
  const string::size_type strLen = str.length();

  // Calculate starting height using dummy text
  int         dr, as, ds;
  XCharStruct cs;
  XTextExtents(*_tX11CS.ptFont, _DTM, strlen(_DTM), &dr, &as, &ds, &cs);

  // Print all the text lines
  string::size_type startIndex = 0;
  string::size_type endIndex   = 0;
  while(startIndex < strLen) {
    // Get the current line text
    string subStr;
    if((endIndex = str.find('\n', startIndex)) == string::npos) {
      subStr     = str.substr(startIndex, strLen - startIndex);
      startIndex = strLen;
    }
    else {
      subStr     = str.substr(startIndex, endIndex - startIndex);
      startIndex = endIndex + 1;
    }
    // Print the text
    if(!subStr.empty()) {
      XTextExtents(*_tX11CS.ptFont, subStr.data(), subStr.length(), &dr, &as, &ds, &cs);
      XDrawString(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                  n32X, n32Y, subStr.data(), subStr.length());
      n32Y += (as + ds);
    }
    else {
      n32Y += ((as + ds) / 2);
    }
  }
#else
  // Preparations
  const string           &str        = csText;
        string            subStr;
  const uint32_t          yInc       = ((*_tX11Style.ptFont)->ascent + (*_tX11Style.ptFont)->descent);
  const uint32_t          yInc2      = yInc / 2;
  const string::size_type strLen     = str.length();
        string::size_type startIndex = 0;
        string::size_type endIndex   = 0;

  // Print all the text lines
  while(startIndex < strLen) {
    // Get the current line text
    if((endIndex = str.find('\n', startIndex)) == string::npos) {
      subStr     = str.substr(startIndex, strLen - startIndex);
      startIndex = strLen;
    }
    else {
      subStr     = str.substr(startIndex, endIndex - startIndex);
      startIndex = endIndex + 1;
    }
    // Print the text
    if(!subStr.empty()) {
      XDrawString(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                  n32X, n32Y, subStr.data(), subStr.length());
      n32Y += yInc;
    }
    else {
      n32Y += yInc2;
    }
  }
#endif
}

void CVideoEnv::DrawUtf8Text(int32_t n32X, int32_t n32Y, const string &csText, ETextAlignment eAlignment)
{
  // Calculate the text width
  XRectangle tl;
  if(eAlignment != ETA_Left)
    Xutf8TextExtents(*_tX11CS.ptFontSet, csText.data(), csText.length(), NULL, &tl);

  // Center on screen horizontally
  int32_t pX;
  if(eAlignment == ETA_CenterScreen)
    pX = (_tX11.un32Width - tl.width) >> 1;
  // Center on the given X coordinate horizontally
  else if(eAlignment == ETA_Center)
    pX = n32X - (tl.width >> 1);
  // Right aligned on the given X coordinate horizontally
  else if(eAlignment == ETA_Right)
    pX = n32X - tl.width;
  // Left aligned on the given X coordinate horizontally
  else
    pX = n32X;

  // Draw the text
  Xutf8DrawString(_tX11.ptDisplay, _tX11.tMainPixmap, *_tX11CS.ptFontSet,
                  _tX11.tMainGC, pX, n32Y, csText.data(), csText.length());

  // Restore the font since Xutf8DrawString() alter the GC font setting
  XSetFont(_tX11.ptDisplay, _tX11.tMainGC, (*_tX11CS.ptFont)->fid);
}

void CVideoEnv::DrawUtf8TextML(int32_t n32X, int32_t n32Y, const string &csText)
{
  // For convenience
  const string           &str    = csText;
  const string::size_type strLen = str.length();

  // Calculate starting height using dummy text
  XRectangle tl;
  Xutf8TextExtents(*_tX11CS.ptFontSet, _DTM, strlen(_DTM), NULL, &tl);

  // Print all the text lines
  string::size_type startIndex = 0;
  string::size_type endIndex   = 0;
  while(startIndex < strLen) {
    // Get the current line text
    string subStr;
    if((endIndex = str.find('\n', startIndex)) == string::npos) {
      subStr     = str.substr(startIndex, strLen - startIndex);
      startIndex = strLen;
    }
    else {
      subStr     = str.substr(startIndex, endIndex - startIndex);
      startIndex = endIndex + 1;
    }
    // Print the text
    if(!subStr.empty()) {
      Xutf8TextExtents(*_tX11CS.ptFontSet, subStr.data(), subStr.length(), NULL, &tl);
      Xutf8DrawString(_tX11.ptDisplay, _tX11.tMainPixmap, *_tX11CS.ptFontSet,
                      _tX11.tMainGC, n32X, n32Y, subStr.data(), subStr.length());
      n32Y += tl.height;
    }
    else {
      n32Y += (tl.height / 2);
    }
  }

  // Restore the font since Xutf8DrawString() alter the GC font setting
  XSetFont(_tX11.ptDisplay, _tX11.tMainGC, (*_tX11CS.ptFont)->fid);
}

void CVideoEnv::ClearMainPixmap(XColor24 xColor)
{
  SetForegroundColor(xColor);
  SetFillStyle(FILL_SOLID);
  XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                 0, 0, _tX11.un32Width, _tX11.un32Height);
  SetFillStyle(_tX11Style.xFillStyle);
}

void CVideoEnv::ApplyBlackScreen()
{
  SetForegroundColor(_tX11.un32BlackPixel);
  SetFillStyle(FILL_STIPPLED);
  XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                 0, 0, _tX11.un32Width, _tX11.un32Height);
  SetFillStyle(_tX11Style.xFillStyle);
}

void CVideoEnv::ApplyWhiteScreen()
{
  SetForegroundColor(_tX11.un32WhitePixel);
  SetFillStyle(FILL_STIPPLED);
  XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                 0, 0, _tX11.un32Width, _tX11.un32Height);
  SetFillStyle(_tX11Style.xFillStyle);
}

void CVideoEnv::DrawGradientBox(uint32_t un32X, uint32_t un32Y, uint32_t un32Width, uint32_t un32Height, XColor24 xLight, XColor24 xDark, XColor24 xFace)
{
  // Calculate the gradient color interpolators
  uint8_t un8L[3], un8F[3], un8D[3];
  ExtractColor(un8L[0], un8L[1], un8L[2], xLight);
  ExtractColor(un8D[0], un8D[1], un8D[2], xDark);
  float fS[3], fE[3], fD[3];
  if(xFace == INVALID_COLOR24_VALUE) {
    for(size_t i = 0; i < 3; ++i) {
      fS[i] = un8L[i];
      fE[i] = un8D[i];
      fD[i] = (fE[i] - fS[i]) / float(un32Height);
    }
  }
  else {
    ExtractColor(un8F[0], un8F[1], un8F[2], xFace);
    for(size_t i = 0; i < 3; ++i) {
      fS[i] = (un8L[i] * 0.75f) + (un8F[i] * 0.25f);
      fE[i] = (un8D[i] * 0.75f) + (un8F[i] * 0.25f);
      fD[i] = (fE[i] - fS[i]) / float(un32Height);
    }
  }

#if 1
  // Draw the gradient block by block
  size_t oy = un32Y;
  size_t ny = oy;
  XColor24 oc = CreateColor(uint8_t(fS[0]), uint8_t(fS[1]), uint8_t(fS[2]));
  XColor24 nc = oc;
  SetFillStyle(FILL_SOLID);
  for(; ny < size_t(un32Y + un32Height); ny++) {
    nc = CreateColor(uint8_t(fS[0]), uint8_t(fS[1]), uint8_t(fS[2]));
    if(oc != nc) {
      XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, oc);
      XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                     un32X, oy, un32Width, ny - oy + 1);
      oc = nc; oy = ny;
    }
    fS[0] += fD[0]; fS[1] += fD[1]; fS[2] += fD[2];
  }
  if(oy != ny-1) {
    XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, nc);
    XFillRectangle(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC,
                   un32X, oy, un32Width, ny - oy);
  }
  SetFillStyle(_tX11Style.xFillStyle);
  _tX11CS.xForeground = nc;
#else
  // Draw the gradient line by line
  const size_t un32X1 = un32X;
  const size_t un32X2 = un32X + un32Width - 1;
  for(size_t un32Yc = un32Y; un32Yc < size_t(un32Y + un32Height); un32Yc++) {
    _tX11CS.xForeground = CreateColor(uint8_t(fS[0]), uint8_t(fS[1]), uint8_t(fS[2]));
    XSetForeground(_tX11.ptDisplay, _tX11.tMainGC, _tX11CS.xForeground);
    XDrawLine(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC, un32X1, un32Yc, un32X2, un32Yc);
    fS[0] += fD[0]; fS[1] += fD[1]; fS[2] += fD[2];
  }
#endif
}

void CVideoEnv::PutRenderImageOnDisplay(int32_t n32X, int32_t n32Y)
{
  (this->*_ConvertRB2RI_)();
#ifdef __USE_XSHM__
  XShmPutImage(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC, _tRI.ptImage,
               0, 0, n32X, n32Y, _tRI.un32Width, _tRI.un32Height, False);
#else
  XPutImage(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainGC, _tRI.ptImage,
             0, 0, n32X, n32Y, _tRI.un32Width, _tRI.un32Height);
#endif
}

void CVideoEnv::UpdateDisplay()
{
  XCopyArea(_tX11.ptDisplay, _tX11.tMainPixmap, _tX11.tMainWindow, _tX11.tMainGC,
            0, 0, _tX11.un32Width, _tX11.un32Height, 0, 0);
  XSync(_tX11.ptDisplay, False);
}

void CVideoEnv::UseDefaultArrowCursor()
{
  XDefineCursor(_tX11.ptDisplay, _tX11.tMainWindow, _tX11.tDefCurArrow);
  XSync(_tX11.ptDisplay, False);
}

void CVideoEnv::UseDefaultWaitCursor()
{
  XDefineCursor(_tX11.ptDisplay, _tX11.tMainWindow, _tX11.tDefCurWait);
  XSync(_tX11.ptDisplay, False);
}


} // namespace vge
